import tkinter as tk
from typing import Callable, Optional

import numpy as np
import cv2

from .multi_slider_widget import MultiSliderOverlay, MultiSliderState

Vector3Di = tuple[int, int, int]
# A color space is a list of 3d vectors, the sum of which is 255 for each component
ColorSpace = list[Vector3Di]
CallbackColorSpace = Callable[[ColorSpace], None]


class ColorSpaceWidget:
    colors = {
        "grey": "#%02x%02x%02x" % (191, 191, 191),
    }

    @staticmethod
    def rgb_segments_to_color_space(red_segments: list[int], green_segments: list[int],
                                    blue_segements: list[int]) -> ColorSpace:
        red_segments = list(sorted(red_segments)) + [255]
        green_segments = list(sorted(green_segments)) + [255]
        blue_segements = list(sorted(blue_segements)) + [255]

        color_space = []
        color_prev = np.array([0, 0, 0], dtype=int)
        for k in range(len(red_segments)):
            rgb = np.array([
                red_segments[k],
                green_segments[k],
                blue_segements[k]
            ], dtype=int)
            color_k = rgb - color_prev
            color_space.append(tuple(color_k.tolist()))
            color_prev = rgb

        return color_space

    @staticmethod
    def color_space_to_rgb_segments(color_space: ColorSpace) -> tuple[list[int], list[int], list[int]]:
        """
        Convert a color space (list of RGB deltas) to RGB segment positions.
        This is the inverse operation of rgb_segments_to_color_space.

        Args:
            color_space: List of RGB delta vectors

        Returns:
            Tuple of (red_segments, green_segments, blue_segments)
        """
        red_segments = []
        green_segments = []
        blue_segments = []

        color_accumulated = np.array([0, 0, 0], dtype=int)
        for delta_r, delta_g, delta_b in color_space:
            delta = np.array([delta_r, delta_g, delta_b], dtype=int)
            color_accumulated = color_accumulated + delta
            red_segments.append(int(color_accumulated[0]))
            green_segments.append(int(color_accumulated[1]))
            blue_segments.append(int(color_accumulated[2]))

        # Get rid of last segment and ensure it is 255
        assert red_segments.pop(-1) == 255
        assert green_segments.pop(-1) == 255
        assert blue_segments.pop(-1) == 255

        return red_segments, green_segments, blue_segments

    def __init__(self,
                 tk_frame: tk.Frame,
                 color_space_name: str,
                 initial_color_space: ColorSpace,
                 on_drag: Optional[CallbackColorSpace] = None,
                 on_release: Optional[CallbackColorSpace] = None,
                 widget_color=None):
        """
        Widget made of 3 multi-sliders (one per channel) to represent a color space
        Creates 3 canvas for sliders, and 1 canvas to display the resulting color (overlaping circles)
        """
        self.on_drag = on_drag
        self.on_release = on_release

        assert len(initial_color_space) > 1

        self.color_space: ColorSpace = []
        for rgb_vector in initial_color_space:
            r, g, b = rgb_vector
            self.color_space.append((int(r), int(g), int(b)))

        self.M = len(initial_color_space)
        self.multi_slider_overlays: list[MultiSliderOverlay] = []

        initial_rgb_segments = ColorSpaceWidget.color_space_to_rgb_segments(self.color_space)

        frame_top = tk.Frame(tk_frame, bg=widget_color)
        self.name = tk.StringVar(value=color_space_name)
        tk.Label(master=frame_top, textvariable=self.name, bg=widget_color).pack(padx=2, side=tk.LEFT)
        self.label_txt = tk.StringVar()
        tk.Label(master=frame_top, textvariable=self.label_txt, bg=widget_color).pack(padx=2, side=tk.TOP)

        frame_top.pack(side=tk.TOP, fill=tk.X, expand=1)

        for channel in range(3):
            canvas = tk.Canvas(master=tk_frame, height=21, borderwidth=0, bg=widget_color)
            canvas.pack(side=tk.TOP, fill=tk.X)

            multi_slider_overlay = MultiSliderOverlay(
                canvas=canvas,
                values=range(256),
                initial_positions=initial_rgb_segments[channel],
                on_drag=lambda positions_values, channel_=channel: self.on_drag_callback(channel_, positions_values),  # type: ignore[misc]
                on_release=self.on_release_callback,
            )
            self.multi_slider_overlays.append(multi_slider_overlay)

        self.visual_canvas = tk.Canvas(master=tk_frame, height=200, borderwidth=0, bg=widget_color)
        self.visual_canvas.pack(side=tk.TOP, fill=tk.X, expand=1)
        self.update_visualization()

    def update_visualization(self) -> None:
        """
        Create and display an RGB image in the visual canvas.
        Currently displays a black image with a red circle.
        """
        # Get canvas dimensions
        self.visual_canvas.update_idletasks()
        height = self.visual_canvas.winfo_height()
        width = height
        radius = max(1, (height - 20) // 3)  # Ensure radius is at least 1 for cv2.circle

        def create_image(color: Vector3Di, theta: float) -> np.ndarray[tuple[int, int, int], np.dtype[np.uint8]]:
            # Create a black RGB image using numpy
            img = np.zeros((height, width, 3), dtype=np.uint8)

            # Draw a red circle using cv2
            center = np.array([width / 2, height / 2], dtype=float)
            off_center = radius / 2
            circle_center = center + off_center * np.array([np.cos(theta), np.sin(theta)])
            circle_center = np.round(circle_center).astype(int)

            thickness = -1  # Filled circle

            # Convert color tuple to proper format for cv2.circle
            color_tuple = (int(color[0]), int(color[1]), int(color[2]))
            cv2.circle(img, tuple(circle_center.tolist()), radius, color_tuple, thickness)

            return img

        img_rgb = create_image((0, 0, 0), 0.0)
        N = len(self.color_space)
        for i, color in enumerate(self.color_space):
            theta = 2.0 * np.pi * (i / N)
            img_rgb += create_image(color, theta)

        # Convert to PIL Image and then to PhotoImage for tkinter
        from PIL import Image, ImageTk
        pil_img = Image.fromarray(img_rgb)
        self.photo_img = ImageTk.PhotoImage(image=pil_img, master=self.visual_canvas)

        # Clear canvas and display the image
        self.visual_canvas.delete("all")
        self.visual_canvas.create_image(0, 0, anchor=tk.NW, image=self.photo_img)

    def update_color_space(self) -> None:
        red_position_list = self.multi_slider_overlays[0].get_positions()
        green_position_list = self.multi_slider_overlays[1].get_positions()
        blue_position_list = self.multi_slider_overlays[2].get_positions()

        self.color_space = ColorSpaceWidget.rgb_segments_to_color_space(red_position_list, green_position_list,
                                                                        blue_position_list)

    def get_color_space(self) -> ColorSpace:
        return self.color_space + []

    def on_drag_callback(self, channel: int, positions_values: MultiSliderState) -> None:
        """
        Wrapper for on_drag callback
        This wrapper is called by the MultiSlider instance
        This wrapper calls the user-defined on_drag callback
        """
        self.update_color_space()
        self.update_visualization()
        if self.on_drag is not None:
            self.on_drag(self.color_space)

    def on_release_callback(self, positions_values: MultiSliderState) -> None:
        """
        Wrapper for on_release callback
        This wrapper is called by the MultiSlider instance
        This wrapper calls the user-defined on_release callback
        """
        self.update_color_space()
        self.update_visualization()
        if self.on_release is not None:
            self.on_release(self.color_space)




import tkinter as tk
from tkinter.colorchooser import askcolor
from typing import Callable, Optional

CallbackColorPicker = Callable[[tuple[int, int, int]], None]


def convert_color_tuple3i_to_hexastr(color_rgb: tuple[int, int, int]) -> str:
    """
    >>> convert_color_tuple3i_to_hexastr((0, 0, 0))
    '#000000'
    >>> convert_color_tuple3i_to_hexastr((255, 0, 10))
    '#ff000a'
    """
    assert isinstance(color_rgb, tuple)
    assert len(color_rgb) == 3
    for c in color_rgb:
        assert isinstance(c, int)
    assert min(color_rgb) >= 0
    assert max(color_rgb) <= 255
    r, g, b = color_rgb
    return f"#{r:02x}{g:02x}{b:02x}"


class ColorPickerWidget:
    def __init__(self, tk_frame, name: str, on_change: CallbackColorPicker, initial_color_rgb: Optional[tuple[int, int, int]] = None):
        self.frame = tk.Frame(tk_frame)
        label = tk.Label(self.frame, text=name)
        label.pack(padx=2, side=tk.LEFT, anchor=tk.W)
        self.on_change = on_change

        if initial_color_rgb is None:
            initial_color_rgb = (0, 0, 0)
        assert isinstance(initial_color_rgb, tuple)
        assert len(initial_color_rgb) == 3
        for c in initial_color_rgb:
            assert isinstance(c, int)
        assert min(initial_color_rgb) >= 0
        assert max(initial_color_rgb) <= 255
        self.color_rgb: tuple[int, int, int] = initial_color_rgb

        color_hex = convert_color_tuple3i_to_hexastr(self.color_rgb)
        self.canvas = tk.Canvas(self.frame, bg=color_hex, bd=3, height=10)
        self.button = tk.Button(self.frame, text="Edit", command=self.callback)

        self.canvas.pack(padx=2, side=tk.LEFT, anchor=tk.W)
        self.button.pack(padx=2, side=tk.LEFT, anchor=tk.W)
        self.frame.pack(padx=4, pady=4, side=tk.TOP, fill=tk.X, expand=1)

    def callback(self):
        # askcolor expects hex string, not RGB tuple
        hex_color = convert_color_tuple3i_to_hexastr(self.get_current_value())
        colors = askcolor(title="Color Chooser", color=hex_color)
        if colors[0] is not None:
            r, g, b = colors[0]
            self.color_rgb = (r, g, b)
        self.canvas["bg"] = convert_color_tuple3i_to_hexastr(self.color_rgb)
        self.on_change(self.color_rgb)

    def get_current_value(self) -> tuple[int, int, int]:
        return self.color_rgb



import numpy as np
import numpy.typing as npt
import cv2
import guibbon as gbn
import tkinter as tk

Image_t = npt.NDArray[np.uint8]


class MyCustomWidget(gbn.WidgetInterface):
    def __init__(self, master: tk.Frame, text: str, on_click_no, on_clik_yes):
        tk.Label(master, text=text).pack(padx=2, side=tk.TOP)

        frame = tk.Frame(master)
        tk.Button(frame, text="No", command=on_click_no).pack(side=tk.LEFT)
        tk.Button(frame, text="Yes", command=on_clik_yes).pack(side=tk.LEFT)
        frame.pack(padx=4, pady=4, side=tk.TOP, fill=tk.BOTH)


class DemoMVCApp:
    def __init__(self, filename):
        img = cv2.imread(filename)
        if img is None:
            raise FileNotFoundError(f"Image file not found: {filename}")
        self.img: Image_t = img.astype(np.uint8)
        self.winname = "demo app"
        gbn.namedWindow(self.winname)
        gbn.createInteractivePoint(self.winname, (100, 100), "point", on_drag=self.on_drag)

        slider = gbn.create_slider(self.winname, "slider", [0, 10, 20, 30], lambda index, val: print("slider", index, val), 2)
        my_widget = gbn.create_custom_widget(self.winname, MyCustomWidget, "Are you all right?", lambda: print("no"), lambda: print("yes"))
        print(my_widget)
        slider.set_position(1)

        gbn.createTrackbar("trackbar", self.winname, 2, 4, lambda val: print("trackbar", val))
        gbn.setTrackbarPos("trackbar", self.winname, 1)
        gbn.setTrackbarPos("trackbar", self.winname, 2)

        gbn.create_button(self.winname, "+2", self.add2)

        self.x = 0
        self.y = 0

        self.res: Image_t
        self.is_update_needed: bool = True

    def update(self):
        w, h = self.img.shape[:2]
        x = self.x / w
        y = self.y / h
        gain_b = max(0.5, min(2.0, 1.5 * x + 0.5))
        gain_r = max(0.5, min(2.0, 1.5 * y + 0.5))

        res = self.img.copy()
        res = res.astype(np.uint16)
        res[:, :, 0:1] = res[:, :, 0:1] * gain_b
        res[:, :, 2:3] = res[:, :, 2:3] * gain_r
        res[res > 255] = 255
        res = res.astype(np.uint8)
        self.res = res
        self.is_update_needed = False

    def show(self):
        while True:
            if self.is_update_needed:
                self.update()
                gbn.imshow(self.winname, self.res)
                self.is_update_needed = False
            gbn.waitKeyEx(20)

    def on_drag(self, event):
        self.x = event.x
        self.y = event.y
        self.is_update_needed = True

    def add2(self):
        slider = gbn.get_slider_instance(self.winname, "slider")
        slider.set_values(slider.get_values() + [100])

        gbn.setTrackbarMin("trackbar", self.winname, 2)


if __name__ == "__main__":
    mc = DemoMVCApp("ressources/dog.jpg")
    mc.show()



import unittest
import tkinter as tk
from unittest.mock import Mock

from guibbon.widgets.check_button_widget import CheckButtonWidget


# Shared Tk root for all tests to avoid tkinter instability
_tk_root = None


def setUpModule():
    """Create a single shared Tk root for all tests in this module"""
    global _tk_root
    _tk_root = tk.Tk()
    _tk_root.withdraw()  # Hide the window


def tearDownModule():
    """Destroy the shared Tk root after all tests"""
    global _tk_root
    if _tk_root:
        try:
            _tk_root.destroy()
        except tk.TclError:
            pass
        _tk_root = None


class TestCheckButtonWidget(unittest.TestCase):
    """Test suite for CheckButtonWidget"""

    def setUp(self) -> None:
        """Set up test fixtures"""
        # Use a Frame as child of the shared root instead of creating new Tk()
        assert _tk_root is not None
        self.frame = tk.Frame(_tk_root, width=400, height=300)
        self.frame.pack(expand=True, fill='both')
        # Update to ensure frame is rendered and has dimensions
        _tk_root.update_idletasks()
        _tk_root.update()

    def tearDown(self) -> None:
        """Clean up after tests"""
        try:
            self.frame.destroy()
        except tk.TclError:
            pass

    def test_create_check_button_widget_default(self) -> None:
        """Test basic CheckButtonWidget creation with default value (False)"""
        callback = Mock()
        widget = CheckButtonWidget(
            tk_frame=self.frame,
            name="Test CheckButton",
            on_change=callback,
        )

        # Check widget was created
        self.assertIsNotNone(widget)
        self.assertIsNotNone(widget.frame)
        self.assertIsNotNone(widget.var)
        
        # Check default value is False
        self.assertEqual(widget.get_current_value(), False)
        self.assertEqual(widget.var.get(), False)

    def test_create_with_initial_value_true(self) -> None:
        """Test creating widget with initial value True"""
        callback = Mock()
        widget = CheckButtonWidget(
            tk_frame=self.frame,
            name="Test CheckButton",
            on_change=callback,
            initial_value=True,
        )

        # Check initial value is True
        self.assertEqual(widget.get_current_value(), True)
        self.assertEqual(widget.var.get(), True)

    def test_create_with_initial_value_false(self) -> None:
        """Test creating widget with initial value False"""
        callback = Mock()
        widget = CheckButtonWidget(
            tk_frame=self.frame,
            name="Test CheckButton",
            on_change=callback,
            initial_value=False,
        )

        # Check initial value is False
        self.assertEqual(widget.get_current_value(), False)
        self.assertEqual(widget.var.get(), False)

    def test_widget_name_stored(self) -> None:
        """Test that widget stores its name"""
        callback = Mock()
        widget_name = "My CheckButton"
        
        widget = CheckButtonWidget(
            tk_frame=self.frame,
            name=widget_name,
            on_change=callback,
        )

        self.assertEqual(widget.name, widget_name)

    def test_get_current_value_false(self) -> None:
        """Test get_current_value returns False when unchecked"""
        callback = Mock()
        widget = CheckButtonWidget(
            tk_frame=self.frame,
            name="Test",
            on_change=callback,
            initial_value=False,
        )

        value = widget.get_current_value()
        self.assertIsInstance(value, bool)
        self.assertEqual(value, False)

    def test_get_current_value_true(self) -> None:
        """Test get_current_value returns True when checked"""
        callback = Mock()
        widget = CheckButtonWidget(
            tk_frame=self.frame,
            name="Test",
            on_change=callback,
            initial_value=True,
        )

        value = widget.get_current_value()
        self.assertIsInstance(value, bool)
        self.assertEqual(value, True)

    def test_callback_on_check(self) -> None:
        """Test callback is called with True when checkbox is checked"""
        callback = Mock()
        widget = CheckButtonWidget(
            tk_frame=self.frame,
            name="Test",
            on_change=callback,
            initial_value=False,
        )

        # Simulate checking the checkbox
        widget.var.set(True)
        widget.callback()

        # Check callback was called with True
        callback.assert_called_once_with(True)

    def test_callback_on_uncheck(self) -> None:
        """Test callback is called with False when checkbox is unchecked"""
        callback = Mock()
        widget = CheckButtonWidget(
            tk_frame=self.frame,
            name="Test",
            on_change=callback,
            initial_value=True,
        )

        # Simulate unchecking the checkbox
        widget.var.set(False)
        widget.callback()

        # Check callback was called with False
        callback.assert_called_once_with(False)

    def test_multiple_state_changes(self) -> None:
        """Test multiple state changes call callback correctly"""
        callback = Mock()
        widget = CheckButtonWidget(
            tk_frame=self.frame,
            name="Test",
            on_change=callback,
            initial_value=False,
        )

        # First change: False -> True
        widget.var.set(True)
        widget.callback()
        self.assertEqual(widget.get_current_value(), True)

        # Second change: True -> False
        widget.var.set(False)
        widget.callback()
        self.assertEqual(widget.get_current_value(), False)

        # Third change: False -> True
        widget.var.set(True)
        widget.callback()
        self.assertEqual(widget.get_current_value(), True)

        # Verify callback was called 3 times
        self.assertEqual(callback.call_count, 3)
        
        # Verify the calls were made with correct values
        calls = [call[0][0] for call in callback.call_args_list]
        self.assertEqual(calls, [True, False, True])

    def test_state_consistency(self) -> None:
        """Test that var.get() and get_current_value() are consistent"""
        callback = Mock()
        widget = CheckButtonWidget(
            tk_frame=self.frame,
            name="Test",
            on_change=callback,
        )

        # Initially False
        self.assertEqual(widget.var.get(), widget.get_current_value())

        # Change to True
        widget.var.set(True)
        self.assertEqual(widget.var.get(), widget.get_current_value())

        # Change back to False
        widget.var.set(False)
        self.assertEqual(widget.var.get(), widget.get_current_value())

    def test_callback_not_called_on_creation(self) -> None:
        """Test that callback is not called during widget creation"""
        callback = Mock()
        
        _ = CheckButtonWidget(
            tk_frame=self.frame,
            name="Test",
            on_change=callback,
            initial_value=True,
        )

        # Callback should not have been called yet
        callback.assert_not_called()

    def test_boolean_var_type(self) -> None:
        """Test that the internal variable is a BooleanVar"""
        callback = Mock()
        widget = CheckButtonWidget(
            tk_frame=self.frame,
            name="Test",
            on_change=callback,
        )

        self.assertIsInstance(widget.var, tk.BooleanVar)

    def test_frame_packing(self) -> None:
        """Test that widget frame is properly packed"""
        callback = Mock()
        widget = CheckButtonWidget(
            tk_frame=self.frame,
            name="Test",
            on_change=callback,
        )

        # Check that the frame has been packed (it should be a child of self.frame)
        children = self.frame.winfo_children()
        self.assertIn(widget.frame, children)

    def test_direct_var_manipulation(self) -> None:
        """Test that directly manipulating var reflects in get_current_value"""
        callback = Mock()
        widget = CheckButtonWidget(
            tk_frame=self.frame,
            name="Test",
            on_change=callback,
            initial_value=False,
        )

        # Directly set the variable
        assert _tk_root is not None
        widget.var.set(True)
        _tk_root.update_idletasks()
        
        # Check it's reflected in get_current_value
        self.assertEqual(widget.get_current_value(), True)

        # Set it back
        widget.var.set(False)
        _tk_root.update_idletasks()
        
        self.assertEqual(widget.get_current_value(), False)

    def test_callback_with_alternating_values(self) -> None:
        """Test callback receives correct alternating boolean values"""
        callback = Mock()
        widget = CheckButtonWidget(
            tk_frame=self.frame,
            name="Test",
            on_change=callback,
            initial_value=False,
        )

        # Simulate several state changes
        for expected_value in [True, False, True, False, True]:
            widget.var.set(expected_value)
            widget.callback()
            
            # Verify last call had expected value
            callback.assert_called_with(expected_value)

        # Verify all calls
        self.assertEqual(callback.call_count, 5)

    def test_widget_name_empty_string(self) -> None:
        """Test creating widget with empty string name"""
        callback = Mock()
        widget = CheckButtonWidget(
            tk_frame=self.frame,
            name="",
            on_change=callback,
        )

        self.assertEqual(widget.name, "")
        self.assertIsNotNone(widget.frame)

    def test_widget_name_special_characters(self) -> None:
        """Test creating widget with special characters in name"""
        callback = Mock()
        special_name = "Test @ CheckButton #123 & More!"
        
        widget = CheckButtonWidget(
            tk_frame=self.frame,
            name=special_name,
            on_change=callback,
        )

        self.assertEqual(widget.name, special_name)


if __name__ == "__main__":
    unittest.main()



import os
import re
import time
import tkinter as tk
import PIL
from typing import Optional, Type, Sequence, Any

import cv2

from .colors import COLORS
from .image_viewer import ImageViewer, MODE
from .keyboard_event_handler import KeyboardEventHandler
from .typedef import Image_t, CallbackPoint, CallbackPolygon, CallbackRect, Point2DList, InteractivePolygon, CallbackMouse
from .typedef import Point2D as Point2D
from .widgets.button_widget import ButtonWidget, CallbackButton
from .widgets.check_button_list_widget import CheckButtonListWidget, CallbackCheckButtonList
from .widgets.check_button_widget import CheckButtonWidget, CallbackCheckButton
from .widgets.color_picker_widget import ColorPickerWidget, CallbackColorPicker
from .widgets.multi_slider_widget import MultiSliderWidget, CallbackMultiSlider
from .widgets.color_space_widget import ColorSpaceWidget, CallbackColorSpace, ColorSpace
from .widgets.radio_buttons_widget import RadioButtonsWidget, CallbackRadioButtons
from .widgets.slider_widget import SliderWidget, CallbackSlider
from .widgets.treeview_widget import TreeviewWidget, CallbackTreeview, TreeNode
from .widgets.widget import WidgetInterface

__version__ = "0.4.0"


def compute_version_info():
    mtch = re.match(r"(\d+).(\d+).(\d+)((-dev)?)$", __version__)

    if mtch is None:
        return [(0, 0, 0), ""]

    major = mtch.group(1)
    minor = mtch.group(2)
    build_nb = mtch.group(3)

    lastindex: int = 0 if mtch.lastindex is None else mtch.lastindex
    mode = "dev" if "dev" in mtch.group(lastindex) else None

    return [(major, minor, build_nb), mode]


__version_info__ = compute_version_info()


def inject(cv2_package):
    cv2_package.createTrackbar = createTrackbar
    cv2_package.destroyAllWindows = not_implemented_error
    cv2_package.destroyWindow = not_implemented_error
    cv2_package.getMouseWheelDelta = not_implemented_error
    cv2_package.getTrackbarPos = getTrackbarPos
    cv2_package.getWindowImageRect = not_implemented_error
    cv2_package.getWindowProperty = getWindowProperty
    cv2_package.imshow = imshow
    cv2_package.moveWindow = not_implemented_error
    cv2_package.namedWindow = namedWindow
    cv2_package.pollKey = not_implemented_error
    cv2_package.resizeWindow = not_implemented_error
    cv2_package.selectROI = not_implemented_error
    cv2_package.selectROIs = not_implemented_error
    cv2_package.setMouseCallback = setMouseCallback
    cv2_package.setTrackbarMax = setTrackbarMax
    cv2_package.setTrackbarMin = setTrackbarMin
    cv2_package.setTrackbarPos = setTrackbarPos
    cv2_package.setWindowProperty = not_implemented_error
    cv2_package.setWindowTitle = not_implemented_error
    cv2_package.startWindowThread = not_implemented_error
    cv2_package.waitKey = waitKey
    cv2_package.waitKeyEx = waitKeyEx


def not_implemented_error():
    raise NotImplementedError("Function not implemented in current version of Guibbon")


def imshow(winname: str, mat: Image_t, mode: MODE = MODE.FIT, cv2_interpolation: Optional[int] = None):
    Guibbon.get_instance(winname).imshow(mat, mode, cv2_interpolation)


def getWindowProperty(winname: str, prop_id: int) -> float:
    if not Guibbon.is_instance(winname):
        return 0.0
    return Guibbon.get_instance(winname).getWindowProperty(prop_id)


def waitKey(delay: int, track_keypress=True, track_keyrelease=False, SilentWarning=False) -> int:
    if not SilentWarning:
        print(
            "WARNING: waitKey function not implemented in Guibbon. The current version is similar to waitKeyEx function. (To suppress this warning, use "
            "SilentWarning=True)"
        )
    return waitKeyEx(delay, track_keypress, track_keyrelease)


def waitKeyEx(delay: int, track_keypress=True, track_keyrelease=False) -> int:
    return Guibbon.waitKeyEx(delay, track_keypress, track_keyrelease)


def setMouseCallback(winname: str, onMouse: CallbackMouse, userdata=None):
    Guibbon.get_instance(winname).setMouseCallback(onMouse, userdata=userdata)


def create_button(winname, text: str, on_click: CallbackButton) -> ButtonWidget:
    return Guibbon.get_instance(winname).create_button(text, on_click)


def get_button_instance(winname: str, text: str) -> ButtonWidget:
    return Guibbon.get_instance(winname).get_button_instance(text)


def create_custom_widget(winname, CustomWidgetClass: Type[WidgetInterface], *params) -> WidgetInterface:
    widget_instance = Guibbon.get_instance(winname).create_custom_widget(CustomWidgetClass, *params)
    return widget_instance


def create_slider(winname: str, slider_name: str, values: Sequence[Any], on_change: CallbackSlider, initial_index: int = 0) -> SliderWidget:
    slider_instance: SliderWidget = Guibbon.get_instance(winname).create_slider(slider_name, values, on_change, initial_index)
    return slider_instance

def create_color_space_widget(winname: str, color_space_name: str, initial_color_space: ColorSpace, on_drag: Optional[CallbackColorSpace] = None,
                              on_release: Optional[CallbackColorSpace] = None) -> ColorSpaceWidget:
    color_space_widget: ColorSpaceWidget = Guibbon.get_instance(winname).create_color_space_widget(color_space_name, initial_color_space, on_drag, on_release)
    return color_space_widget


def create_multislider(winname: str, multislider_name: str, values: Sequence[Any], initial_indexes: Sequence[int], on_drag: Optional[CallbackMultiSlider] = None,
                       on_release: Optional[CallbackMultiSlider] = None) -> MultiSliderWidget:
    multislider_instance: MultiSliderWidget = Guibbon.get_instance(winname).create_multislider(multislider_name, values, initial_indexes, on_drag, on_release)
    return multislider_instance


def get_slider_instance(winname: str, slider_name: str) -> SliderWidget:
    slider_instance: SliderWidget = Guibbon.get_instance(winname).get_slider_instance(slider_name)
    return slider_instance


def createTrackbar(trackbarName, windowName, value, count, onChange):
    """
    Same trackbar as in opencv. However, for a more pythonic signature, you may want to use create_slider(...) instead
    """
    values = list(range(count + 1))
    initial_index = value

    def on_change(index, val):
        return onChange(val)

    Guibbon.get_instance(windowName).create_slider(trackbarName, values, on_change, initial_index)


def setTrackbarPos(trackbarname, winname, pos):
    trackbar_instance = get_slider_instance(winname, trackbarname)
    trackbar_instance.set_position(pos)


def getTrackbarPos(trackbarname, winname):
    trackbar_instance = get_slider_instance(winname, trackbarname)
    return trackbar_instance.get_position()


def setTrackbarMin(trackbarname, winname, minval):
    trackbar_instance = get_slider_instance(winname, trackbarname)
    current_minval = trackbar_instance.get_values()[0]
    maxval = trackbar_instance.get_values()[-1]
    values = list(range(minval, maxval + 1))
    new_index = max(trackbar_instance.get_position() + current_minval - minval, 0)
    trackbar_instance.set_values(values, new_index)


def getTrackbarMin(trackbarname, winname):
    trackbar_instance = get_slider_instance(winname, trackbarname)
    return trackbar_instance.get_values()[0]


def setTrackbarMax(trackbarname, winname, maxval):
    trackbar_instance = get_slider_instance(winname, trackbarname)
    minval = trackbar_instance.get_values()[0]
    values = list(range(minval, maxval + 1))
    trackbar_instance.set_values(values)


def getTrackbarMax(trackbarname, winname):
    trackbar_instance = get_slider_instance(winname, trackbarname)
    return trackbar_instance.get_values()[-1]


def namedWindow(winname):
    Guibbon.get_instance(winname)


def create_window(winname: str) -> "Guibbon":
    return Guibbon.get_instance(winname)


def create_radio_buttons(winname: str, name: str, options: list[str], on_change: CallbackRadioButtons) -> RadioButtonsWidget:
    return Guibbon.get_instance(winname).create_radio_buttons(name, options, on_change)


def get_radio_buttons(winname: str, name: str) -> RadioButtonsWidget:
    return Guibbon.get_instance(winname).get_radio_buttons_instance(name)


def create_check_button(winname: str, name: str, on_change: CallbackCheckButton, initial_value: bool = False) -> CheckButtonWidget:
    return Guibbon.get_instance(winname).create_check_button(name, on_change, initial_value)


def create_check_button_list(winname: str, name: str, options: list[str], on_change: CallbackCheckButtonList, initial_values: Optional[list[bool]] = None) -> CheckButtonListWidget:
    return Guibbon.get_instance(winname).create_check_button_list(name, options, on_change, initial_values)


def create_color_picker(winname: str, name: str, on_change: CallbackColorPicker, initial_color_rgb: Optional[tuple[int, int, int]] = None) -> ColorPickerWidget:
    return Guibbon.get_instance(winname).create_color_picker(name, on_change, initial_color_rgb)


def create_treeview(winname: str, name: str, tree: TreeNode, on_click: CallbackTreeview) -> TreeviewWidget:
    return Guibbon.get_instance(winname).create_treeview(name, tree, on_click)


def createInteractivePoint(
        windowName,
        point_xy,
        label="",
        on_click: CallbackPoint = None,
        on_drag: CallbackPoint = None,
        on_release: CallbackPoint = None,
        magnet_points: Optional[Point2DList] = None,
):
    Guibbon.get_instance(windowName).image_viewer.createInteractivePoint(point_xy, label, on_click, on_drag, on_release, magnet_points)


def createInteractivePolygon(
        windowName,
        point_xy_list,
        label="",
        on_click: CallbackPolygon = None,
        on_drag: CallbackPolygon = None,
        on_release: CallbackPolygon = None,
        magnet_points: Optional[Point2DList] = None,
) -> InteractivePolygon:
    """
    Draws a polygon over the displayed image. The corners of the polygon can be dragged by the user, triggering a callback.
    """
    ipolygon: InteractivePolygon
    ipolygon = Guibbon.get_instance(windowName).image_viewer.createInteractivePolygon(point_xy_list, label, on_click, on_drag, on_release, magnet_points)
    return ipolygon


def createInteractiveRectangle(
        windowName,
        point0_xy,
        point1_xy,
        label="",
        on_click: CallbackRect = None,
        on_drag: CallbackRect = None,
        on_release: CallbackRect = None,
        magnet_points: Optional[Point2DList] = None,
) -> InteractivePolygon:
    """Create and returns object of type createInteractiveRectangle windows named windowName."""
    irect: InteractivePolygon
    irect = Guibbon.get_instance(windowName).image_viewer.createInteractiveRectangle(point0_xy, point1_xy, label, on_click, on_drag, on_release, magnet_points)
    return irect


class Guibbon:
    """The Guibbon object contains the list of open windows"""

    root: tk.Tk
    is_alive: bool = False
    instances: dict[str, "Guibbon"] = {}
    active_instance_name: Optional[str]
    is_timeout: bool
    keyboard: KeyboardEventHandler

    @staticmethod
    def init():
        Guibbon.root = tk.Tk()
        Guibbon.is_alive = True
        Guibbon.keyboard = KeyboardEventHandler()
        Guibbon.root.withdraw()
        Guibbon.reset()

    @staticmethod
    def reset():
        Guibbon.is_timeout = False

    @staticmethod
    def on_timeout():
        Guibbon.is_timeout = True

    @staticmethod
    def is_instance(winname: str) -> bool:
        assert isinstance(winname, str)
        return winname in Guibbon.instances.keys()

    @staticmethod
    def get_instance(winname) -> "Guibbon":
        assert isinstance(winname, str)
        if winname not in Guibbon.instances.keys():
            Guibbon.instances[winname] = Guibbon(winname)
        Guibbon.active_instance_name = winname
        return Guibbon.instances[winname]

    @staticmethod
    def get_active_instance() -> "Guibbon":
        return Guibbon.get_instance(Guibbon.active_instance_name)

    @staticmethod
    def waitKeyEx(delay, track_keypress=True, track_keyrelease=False) -> int:
        Guibbon.reset()

        if delay > 0:
            Guibbon.get_active_instance().window.after(delay, Guibbon.on_timeout)  # TODO: make threadsafe

        tic = time.time()
        while True:
            Guibbon.root.update_idletasks()
            Guibbon.root.update()  # root can be destroyed at this line

            if not Guibbon.is_alive:
                return -1

            if track_keypress and Guibbon.keyboard.is_keypress_updated:
                Guibbon.keyboard.is_keypress_updated = False
                return Guibbon.keyboard.last_keypressed
            if track_keyrelease and Guibbon.keyboard.is_keyrelease_updated:
                Guibbon.keyboard.is_keyrelease_updated = False
                return Guibbon.keyboard.last_keyreleased
            if Guibbon.is_timeout:
                return -1

            dt = time.time() - tic
            sleep_time = max(0.0, 1 / 20 - dt)
            time.sleep(sleep_time)
            tic = time.time()

    def __init__(self, winname):
        if not Guibbon.is_alive:
            Guibbon.init()

        # Make master root windows invisible
        self.window = tk.Toplevel(Guibbon.root)  # type: ignore
        self.window.protocol("WM_DELETE_WINDOW", self.on_closing)  # add callback when user closes the window
        self.winname = winname
        self.window.title(self.winname)
        iconpath = os.path.join( os.path.dirname(os.path.abspath(__file__)) , "icons", "icon32.png")
        
        # Load icon using PIL and keep a strong reference
        self.icon: Optional[tk.PhotoImage] = None
        try:
            pil_icon = PIL.Image.open(iconpath)
            # Convert to tk.PhotoImage properly
            self.icon = tk.PhotoImage(file=iconpath)
            self.window.iconphoto(False, self.icon)
        except Exception as e:
            # If that fails, try alternative method or skip
            try:
                # Alternative: use PIL ImageTk but ensure it's kept in memory
                pil_icon = PIL.Image.open(iconpath)
                pil_photo = PIL.ImageTk.PhotoImage(pil_icon)
                # Store in root to keep reference alive across all windows
                if not hasattr(Guibbon.root, '_icon_ref'):
                    setattr(Guibbon.root, '_icon_ref', [])
                icon_refs: list[Any] = getattr(Guibbon.root, '_icon_ref')
                icon_refs.append(pil_photo)
                self.window.iconphoto(False, pil_photo)  # type: ignore[arg-type]
            except Exception as e2:
                print(f"Warning: could not load guibbon icon: {e}, {e2}")
                self.icon = None

        self.window.bind("<KeyPress>", Guibbon.keyboard.on_event)
        self.window.bind("<KeyRelease>", Guibbon.keyboard.on_event)

        self.frame = tk.Frame(master=self.window, bg=COLORS.background)

        # Load an image in the script
        self.img_ratio = 4 / 4
        self.image_viewer = ImageViewer(self.frame, height=720, width=int(720 * self.img_ratio))

        # add dummy image to image_viewer
        # img = np.zeros(shape=(100, 100, 3), dtype=np.uint8)
        # self.imshow(img)

        self.ctrl_frame = tk.Frame(master=self.frame, bg=COLORS.ctrl_panel)
        dummy_canvas = tk.Canvas(master=self.ctrl_frame, height=0, width=300)
        dummy_canvas.pack()

        self.sliders_by_names: dict[str, SliderWidget] = {}
        self.custom_widtgets_by_names: dict[str, WidgetInterface] = {}
        self.radio_buttons_by_names: dict[str, RadioButtonsWidget] = {}
        self.buttons_by_names: dict[str, ButtonWidget] = {}

        self.frame.pack()
        self.image_viewer.frame.pack(side=tk.LEFT)
        # self.ctrl_frame.pack_propagate(False)
        self.ctrl_frame.pack(fill=tk.X)

    def on_closing(self):
        print("Destroy Root", self.winname)
        Guibbon.instances.pop(self.winname)
        self.window.destroy()

        if len(Guibbon.instances) == 0:
            print("destroy root master")
            Guibbon.root.destroy()
            Guibbon.is_alive = False
        elif Guibbon.active_instance_name == self.winname:
            Guibbon.active_instance_name = list(Guibbon.instances.keys())[-1]

    def create_button(self, text: str, on_click: CallbackButton) -> ButtonWidget:
        tk_frame = tk.Frame(self.ctrl_frame, bg=COLORS.ctrl_panel)
        tk_frame.pack_propagate(True)
        button = ButtonWidget(tk_frame, text, on_click)
        self.buttons_by_names[text] = button
        tk_frame.pack(padx=4, pady=4, side=tk.TOP, fill=tk.BOTH)
        return button

    def get_button_instance(self, text: str) -> ButtonWidget:
        return self.buttons_by_names[text]

    def create_slider(self, slider_name: str, values: Sequence[Any], on_change: CallbackSlider, initial_index: int = 0):
        tk_frame = tk.Frame(self.ctrl_frame, bg=COLORS.widget)
        slider = SliderWidget(tk_frame, slider_name, values, initial_index, on_change, COLORS.widget)
        self.sliders_by_names[slider_name] = slider
        tk_frame.pack(padx=4, pady=4, side=tk.TOP, fill=tk.X, expand=1)
        return slider

    def get_slider_instance(self, slider_name: str):
        return self.sliders_by_names[slider_name]

    def create_color_space_widget(self, color_space_name: str, initial_color_space: ColorSpace, on_drag: Optional[CallbackColorSpace] = None,
                                  on_release: Optional[CallbackColorSpace] = None) -> ColorSpaceWidget:
        tk_frame = tk.Frame(self.ctrl_frame, bg=COLORS.widget)
        tk_frame.pack(padx=4, pady=4, side=tk.TOP, fill=tk.X, expand=1)
        color_space_widget = ColorSpaceWidget(tk_frame, color_space_name, initial_color_space, on_drag, on_release, COLORS.widget)
        return color_space_widget

    def create_multislider(self, multislider_name: str, values: Sequence[Any], initial_indexes: Sequence[int], on_drag: Optional[CallbackMultiSlider] = None,
                           on_release: Optional[CallbackMultiSlider] = None) -> MultiSliderWidget:
        tk_frame = tk.Frame(self.ctrl_frame, bg=COLORS.widget)
        tk_frame.pack(padx=4, pady=4, side=tk.TOP, fill=tk.X, expand=1)
        multi_slider_widget = MultiSliderWidget(tk_frame, multislider_name, values, initial_indexes, on_drag, on_release, COLORS.widget)
        return multi_slider_widget

    def create_custom_widget(self, CustomWidgetClass: Type[WidgetInterface], *params) -> WidgetInterface:
        tk_frame = tk.Frame(self.ctrl_frame, bg=COLORS.widget)
        widget_instance = CustomWidgetClass(tk_frame, *params)
        tk_frame.pack(padx=4, pady=4, side=tk.TOP, fill=tk.X, expand=1)
        return widget_instance

    def create_radio_buttons(self, name: str, options: list[str], on_change: CallbackRadioButtons) -> RadioButtonsWidget:
        tk_frame = tk.Frame(self.ctrl_frame, bg=COLORS.widget)
        radio_buttons = RadioButtonsWidget(tk_frame, name, options, on_change)
        self.radio_buttons_by_names[name] = radio_buttons
        tk_frame.pack(padx=4, pady=4, side=tk.TOP, fill=tk.X, expand=1)
        return radio_buttons

    def get_radio_buttons_instance(self, name: str) -> RadioButtonsWidget:
        return self.radio_buttons_by_names[name]

    def create_check_button(self, name: str, on_change: CallbackCheckButton, initial_value: bool = False) -> CheckButtonWidget:
        tk_frame = tk.Frame(self.ctrl_frame, bg=COLORS.widget)
        cb = CheckButtonWidget(tk_frame, name, on_change, initial_value)
        tk_frame.pack(padx=4, pady=4, side=tk.TOP, fill=tk.X, expand=1)
        return cb

    def create_check_button_list(self, name: str, options: list[str], on_change: CallbackCheckButtonList, initial_values: Optional[list[bool]] = None) -> CheckButtonListWidget:
        tk_frame = tk.Frame(self.ctrl_frame)
        cb = CheckButtonListWidget(tk_frame, name, options, on_change, initial_values)
        tk_frame.pack(padx=4, pady=4, side=tk.TOP, fill=tk.X, expand=1)
        return cb

    def create_color_picker(self, name: str, on_change: CallbackColorPicker, initial_color_rgb: Optional[tuple[int, int, int]] = None) -> ColorPickerWidget:
        tk_frame = tk.Frame(self.ctrl_frame)
        cpw = ColorPickerWidget(tk_frame, name, on_change, initial_color_rgb)
        tk_frame.pack(padx=4, pady=4, side=tk.TOP, fill=tk.X, expand=1)
        return cpw

    def create_treeview(self, name: str, tree: TreeNode, on_click: CallbackTreeview) -> TreeviewWidget:
        tk_frame = tk.Frame(self.ctrl_frame)
        widget = TreeviewWidget(tk_frame, name, tree, on_click)
        tk_frame.pack(padx=4, pady=4, side=tk.TOP, fill=tk.X, expand=1)
        return widget

    def imshow(self, mat: Image_t, mode: MODE = MODE.FIT, cv2_interpolation: Optional[int] = None):
        self.image_viewer.imshow(mat, mode, cv2_interpolation)

    def getWindowProperty(self, prop_id: int) -> float:
        """
        TODO: not all flags ar handled
            cv2.WND_PROP_FULLSCREEN:    fullscreen property (can be WINDOW_NORMAL or WINDOW_FULLSCREEN).
            cv2.WND_PROP_AUTOSIZE:      autosize property (can be WINDOW_NORMAL or WINDOW_AUTOSIZE).
            cv2.WND_PROP_ASPECT_RATIO:  window's aspect ration (can be set to WINDOW_FREERATIO or WINDOW_KEEPRATIO).
            cv2.WND_PROP_OPENGL:        opengl support.
            cv2.WND_PROP_VISIBLE:       checks whether the window exists and is visible
            cv2.WND_PROP_TOPMOST:       property to toggle normal window being topmost or not
            cv2.WND_PROP_VSYNC:         enable or disable VSYNC (in OpenGL mode)
        """
        if prop_id == cv2.WND_PROP_VISIBLE:
            return 1.0 if self.window.state() == "normal" else 0.0
        else:
            raise NotImplementedError(
                f"Function getWindowProperty is not fully implemented in current version of Guibbon and does not support the provided "
                f"flag: prop_id={prop_id}"
            )

    def setMouseCallback(self, onMouse: CallbackMouse, userdata=None):
        self.image_viewer.setMouseCallback(onMouse, userdata)



import re
import sys
import unittest
from typing import Tuple

import cv2
import numpy as np
import toml

import guibbon as gbn

eps = sys.float_info.epsilon


class TestPackage(unittest.TestCase):
    def test_version_match(self):
        ver0 = gbn.__version__
        toml_data = toml.load("pyproject.toml")
        version_line = toml_data["project"]["version"]
        ver1 = version_line.split()[-1]
        ver1 = ver1.replace(".dev0", "-dev")  # I don't know where the ".dev0" comes from but it corresponds to my "-dev"
        self.assertEqual(ver0, ver1, "Package version (in pyproject.toml) and __version__ (in __init__.py) must match")

    def test_version_format(self):
        mtch = re.match(r"(\d+).(\d+).(\d+)((-dev)?)$", gbn.__version__)
        self.assertIsNotNone(mtch, 'Version number must match regex: ' + r"(\d+).(\d+).(\d+)((-dev)?)$")

    def test_version_info_format(self):
        version_digits, mode = gbn.__version_info__
        version = ".".join([str(digit) for digit in version_digits])
        if mode is not None:
            version += "-" + mode
        self.assertEqual(version, gbn.__version__)


class TestGuibbon(unittest.TestCase):
    def tearDown(self) -> None:
        gbn.Guibbon.instances = {}
        gbn.Guibbon.active_instance_name = None

    def test_package(self):
        self.assertIsNotNone(gbn.__version__)
        self.assertIsNotNone(gbn.__version_info__)

    def test_inject(self):
        cv2_func = cv2.namedWindow
        gbn_func = gbn.namedWindow
        self.assertNotEqual(cv2_func, gbn_func)
        gbn.inject(cv2)
        self.assertEqual(cv2.namedWindow, gbn_func, msg="gbn.inject must correctly inject function to cv2")

    def test_namedWindow(self):
        self.assertEqual(len(gbn.Guibbon.instances), 0, msg="At initialisation, number of instances must be 0.")
        res = gbn.namedWindow("win0")
        self.assertIsNone(res, msg="function gbn.namedWindow(winname) must return None")
        self.assertEqual(len(gbn.Guibbon.instances), 1, msg="After first use of gbn.namedWindow(...), number of instances must be 1.")
        self.assertIsNotNone(gbn.Guibbon.instances["win0"], msg='After use of gbn.namedWindow("win0"), gbn.Guibbon.instances["win0"] must not be None')

    def test_multiple_instances(self):
        self.assertEqual(len(gbn.Guibbon.instances), 0, msg="At initialisation, number of instances must be 0.")
        gbn.namedWindow("win0")
        gbn.namedWindow("win1")
        self.assertEqual(len(gbn.Guibbon.instances), 2, msg="After 2 uses of gbn.namedWindow(...), number of instances must be 2.")

    def test_imshow_and_getWindowProperty(self):
        img = np.zeros((10, 10), dtype=np.uint8)
        winname = "win0"
        self.assertEqual(len(gbn.Guibbon.instances), 0, msg="At initialisation, number of instances must be 0.")
        res = gbn.getWindowProperty(winname, cv2.WND_PROP_VISIBLE)
        self.assertTrue(abs(res - 0.0) <= eps, msg="Non existant windows must have property WND_PROP_VISIBLE set to 0.0")
        self.assertFalse(gbn.Guibbon.is_instance(winname), msg="gbn.Guibbon.is_instance must return false for non-existant windows")
        self.assertEqual(len(gbn.Guibbon.instances), 0, msg="getWindowProperty should not be able to create new window")

        res = gbn.imshow(winname, img)
        self.assertEqual(len(gbn.Guibbon.instances), 1, msg="gbn.imshow must be able to create new window")
        self.assertTrue(gbn.Guibbon.is_instance(winname), msg="gbn.imshow must be able to create new window with given name")
        self.assertIsNone(res, msg="gbn.imshow must return None")

        res = gbn.getWindowProperty(winname, cv2.WND_PROP_VISIBLE)
        self.assertTrue(abs(res - 1.0) <= eps, msg="Existant and visible windows must have property WND_PROP_VISIBLE set to 1.0")
        gbn.Guibbon.instances = {}  # TODO: use
        res = gbn.getWindowProperty(winname, cv2.WND_PROP_VISIBLE)
        self.assertTrue(abs(res - 0.0) <= eps, msg="Deleting instance of a windows must destroy it. getWindowProperty must return 0")


def find_widget_by_name(guibbon_instance, widgetname):
    last_guibbon_widget = list(guibbon_instance.ctrl_frame.children.values())[-1]

    # Not sure if this is good practice because this line depends on how the widget is implemented.
    # Any minor modification of implementation could make this line raise an exception.
    widgetname_list = [key for key in last_guibbon_widget.children.keys() if widgetname in key]
    assert len(widgetname_list) == 1
    widgetname = widgetname_list[0]
    widget = last_guibbon_widget.children[widgetname]
    return widget


class TestGuibbon_checkbutton(unittest.TestCase):
    def setUp(self):
        self.winname = "win0"
        gbn.namedWindow(self.winname)
        self.guibbon_instance = gbn.Guibbon.instances["win0"]
        self.triggered = None
        self.args: Tuple[bool, ...]

    def callback(self, *args: bool):
        print("callback checkbutton triggered", args)
        self.triggered = True
        self.args = args

    def test_create_check_button(self):
        widget = gbn.create_check_button(winname=self.winname, name="CheckButton", on_change=self.callback, initial_value=False)
        self.assertIsNotNone(widget, msg="function gbn.create_check_button must not return None")
        self.assertIsInstance(widget, gbn.CheckButtonWidget, msg="function gbn.create_check_button must return instance of CheckButtonWidget")

        self.assertIsNone(self.triggered)

        widget.on_change(True)
        self.assertTrue(self.triggered)
        self.assertTupleEqual(self.args, (True,))

        widget.on_change(False)
        self.assertTrue(self.triggered)
        self.assertTupleEqual(self.args, (False,))


class TestGuibbon_TrackBar(unittest.TestCase):
    def setUp(self):
        self.winname = "win0"
        gbn.namedWindow(self.winname)
        self.guibbon_instance = gbn.Guibbon.instances["win0"]
        self.triggered = None

    def callback(self, *args):
        print("callback trackBar triggered", args)
        self.triggered = True

    def test_TrackBar(self):
        """
        testing 3 functions:
         - createTrackBar
         - getTrackBarPos
         - setTrackBarPos
        """
        name = "TrackBar"
        res = gbn.createTrackbar(trackbarName=name, windowName=self.winname, value=1, count=10, onChange=self.callback)
        self.assertIsNone(res, msg="function gbn.createTrackBar must return None")

        value = gbn.getTrackbarPos(name, self.winname)
        self.assertEqual(value, 1, msg="function gbn.getTrackbarPos must return correct value")

        self.triggered = False
        gbn.setTrackbarPos(name, self.winname, 8)
        self.assertTrue(self.triggered, "function gbn.setTrackbarPos must trigger callback")
        value = gbn.getTrackbarPos(name, self.winname)
        self.assertEqual(value, 8, msg="function gbn.getTrackbarPos must return correct value")

    def test_TrackBar_min_max(self):
        """
        testing 4 functions:
         - setTrackbarMax
         - setTrackbarMin
         - getTrackbarMax
         - getTrackbarMin
        """
        name = "TrackBar"
        gbn.createTrackbar(trackbarName=name, windowName=self.winname, value=1, count=10, onChange=self.callback)
        initial_min = gbn.getTrackbarMin(trackbarname=name, winname=self.winname)
        initial_max = gbn.getTrackbarMax(trackbarname=name, winname=self.winname)

        self.assertEqual(initial_min, 0)
        self.assertEqual(initial_max, 10)

        gbn.setTrackbarMin(trackbarname=name, winname=self.winname, minval=5)
        gbn.setTrackbarMax(trackbarname=name, winname=self.winname, maxval=20)

        new_min = gbn.getTrackbarMin(trackbarname=name, winname=self.winname)
        new_max = gbn.getTrackbarMax(trackbarname=name, winname=self.winname)

        self.assertEqual(new_min, 5)
        self.assertEqual(new_max, 20)


class TestGuibbon_other(unittest.TestCase):
    def test_not_implemented(self):
        with self.assertRaises(NotImplementedError):
            gbn.not_implemented_error()


if __name__ == "__main__":
    unittest.main()



import tkinter as tk
from typing import Any, Sequence, Optional

from guibbon.interactive_overlays import MultiSliderOverlay, MultiSliderState, CallbackMultiSlider


class MultiSliderWidget:


    def __init__(self,
                 tk_frame: tk.Frame,
                 multi_slider_name: str,
                 values: Sequence[Any],
                 initial_positions: Sequence[int],
                 on_drag: Optional[CallbackMultiSlider] = None,
                 on_release: Optional[CallbackMultiSlider] = None,
                 widget_color=None):

        frame_top = tk.Frame(tk_frame, bg=widget_color)
        self.name = tk.StringVar(value=multi_slider_name)
        tk.Label(master=frame_top, textvariable=self.name, bg=widget_color).pack(padx=2, side=tk.LEFT)
        self.label_txt = tk.StringVar()
        tk.Label(master=frame_top, textvariable=self.label_txt, bg=widget_color).pack(padx=2, side=tk.TOP)

        frame_top.pack(side=tk.TOP, fill=tk.X, expand=1)

        canvas = tk.Canvas(master=tk_frame, height=21, borderwidth=0, bg=widget_color)
        canvas.pack(side=tk.TOP, fill=tk.X)

        self.multi_slider_overlay = MultiSliderOverlay(
            canvas,
            values,
            initial_positions,
            on_drag=self.on_drag_callback,
            on_release=self.on_release_callback,
        )

        self.on_drag = on_drag
        self.on_release = on_release

        self.update_label()


    def update_label(self) -> None:
        values = self.multi_slider_overlay.get_values()
        positions = self.multi_slider_overlay.get_positions()
        pvalues = [str(values[pos]) for pos in positions]
        self.label_txt.set("[" + ", ".join(pvalues) + "]")


    def on_drag_callback(self, positions_values: MultiSliderState) -> None:
        """
        Wrapper for on_drag callback
        This wrapper is called by the MultiSlider instance
        This wrapper calls the user-defined on_drag callback
        """
        self.update_label()
        if self.on_drag is not None:
            self.on_drag(positions_values)

    def on_release_callback(self, positions_values: MultiSliderState) -> None:
        """
        Wrapper for on_release callback
        This wrapper is called by the MultiSlider instance
        This wrapper calls the user-defined on_release callback
        """
        self.update_label()
        if self.on_release is not None:
            self.on_release(positions_values)

    def add_cursor(self, position: int = 0) -> None:
        self.multi_slider_overlay.add_cursor(position)
        self.update_label()

    def remove_cursor(self) -> None:
        self.multi_slider_overlay.remove_cursor()
        self.update_label()

    def get_positions(self):
        return self.multi_slider_overlay.get_positions()

    def set_positions(self, positions: Sequence[int], trigger_callback=True):
        self.multi_slider_overlay.set_positions(positions, trigger_callback)
        self.update_label()

    def get_values(self):
        return self.multi_slider_overlay.get_values()

    def set_values(self, values: Sequence[Any], new_position=None, trigger_callback=True):
        # MultiSliderOverlay.set_values doesn't accept new_position parameter
        self.multi_slider_overlay.set_values(values, trigger_callback)
        self.update_label()





import unittest
import tkinter as tk
from unittest.mock import Mock, patch

from guibbon.widgets.color_picker_widget import (
    ColorPickerWidget,
    convert_color_tuple3i_to_hexastr,
)


# Shared Tk root for all tests to avoid tkinter instability
_tk_root = None


def setUpModule():
    """Create a single shared Tk root for all tests in this module"""
    global _tk_root
    _tk_root = tk.Tk()
    _tk_root.withdraw()  # Hide the window


def tearDownModule():
    """Destroy the shared Tk root after all tests"""
    global _tk_root
    if _tk_root:
        try:
            _tk_root.destroy()
        except tk.TclError:
            pass
        _tk_root = None


class TestColorConversion(unittest.TestCase):
    """Test suite for color conversion utility functions"""

    def test_convert_black(self) -> None:
        """Test converting black color"""
        result = convert_color_tuple3i_to_hexastr((0, 0, 0))
        self.assertEqual(result, "#000000")

    def test_convert_white(self) -> None:
        """Test converting white color"""
        result = convert_color_tuple3i_to_hexastr((255, 255, 255))
        self.assertEqual(result, "#ffffff")

    def test_convert_red(self) -> None:
        """Test converting pure red"""
        result = convert_color_tuple3i_to_hexastr((255, 0, 0))
        self.assertEqual(result, "#ff0000")

    def test_convert_green(self) -> None:
        """Test converting pure green"""
        result = convert_color_tuple3i_to_hexastr((0, 255, 0))
        self.assertEqual(result, "#00ff00")

    def test_convert_blue(self) -> None:
        """Test converting pure blue"""
        result = convert_color_tuple3i_to_hexastr((0, 0, 255))
        self.assertEqual(result, "#0000ff")

    def test_convert_mixed_color(self) -> None:
        """Test converting mixed color with different values"""
        result = convert_color_tuple3i_to_hexastr((123, 45, 67))
        self.assertEqual(result, "#7b2d43")

    def test_convert_single_digit_values(self) -> None:
        """Test that single digit values are zero-padded"""
        result = convert_color_tuple3i_to_hexastr((5, 10, 15))
        self.assertEqual(result, "#050a0f")

    def test_convert_invalid_tuple_length(self) -> None:
        """Test that invalid tuple length raises assertion"""
        with self.assertRaises(AssertionError):
            convert_color_tuple3i_to_hexastr((255, 255))  # type: ignore

    def test_convert_negative_value(self) -> None:
        """Test that negative values raise assertion"""
        with self.assertRaises(AssertionError):
            convert_color_tuple3i_to_hexastr((-1, 0, 0))

    def test_convert_value_too_large(self) -> None:
        """Test that values > 255 raise assertion"""
        with self.assertRaises(AssertionError):
            convert_color_tuple3i_to_hexastr((256, 0, 0))


class TestColorPickerWidget(unittest.TestCase):
    """Test suite for ColorPickerWidget"""

    def setUp(self) -> None:
        """Set up test fixtures"""
        # Use a Frame as child of the shared root instead of creating new Tk()
        assert _tk_root is not None
        self.frame = tk.Frame(_tk_root, width=400, height=300)
        self.frame.pack(expand=True, fill='both')
        # Update to ensure frame is rendered and has dimensions
        _tk_root.update_idletasks()
        _tk_root.update()

    def tearDown(self) -> None:
        """Clean up after tests"""
        try:
            self.frame.destroy()
        except tk.TclError:
            pass

    def test_create_color_picker_widget(self) -> None:
        """Test basic ColorPickerWidget creation with default color"""
        callback = Mock()
        widget = ColorPickerWidget(
            tk_frame=self.frame,
            name="Test Color",
            on_change=callback,
        )

        # Check widget was created
        self.assertIsNotNone(widget)
        self.assertIsNotNone(widget.frame)
        self.assertIsNotNone(widget.canvas)
        self.assertIsNotNone(widget.button)

        # Check default color is black
        self.assertEqual(widget.get_current_value(), (0, 0, 0))
        self.assertEqual(widget.canvas["bg"], "#000000")

    def test_create_with_initial_color(self) -> None:
        """Test creating widget with custom initial color"""
        callback = Mock()
        initial_color = (128, 64, 192)
        
        widget = ColorPickerWidget(
            tk_frame=self.frame,
            name="Custom Color",
            on_change=callback,
            initial_color_rgb=initial_color,
        )

        # Check initial color is set correctly
        self.assertEqual(widget.get_current_value(), initial_color)
        self.assertEqual(widget.canvas["bg"], "#8040c0")

    def test_get_current_value(self) -> None:
        """Test get_current_value returns the correct color"""
        callback = Mock()
        initial_color = (100, 150, 200)
        
        widget = ColorPickerWidget(
            tk_frame=self.frame,
            name="Test",
            on_change=callback,
            initial_color_rgb=initial_color,
        )

        value = widget.get_current_value()
        self.assertEqual(value, initial_color)
        self.assertIsInstance(value, tuple)
        self.assertEqual(len(value), 3)

    def test_callback_with_color_selected(self) -> None:
        """Test callback is called when color is selected"""
        callback = Mock()
        widget = ColorPickerWidget(
            tk_frame=self.frame,
            name="Test",
            on_change=callback,
            initial_color_rgb=(50, 50, 50),
        )

        # Mock askcolor to return a new color (using integers to match widget's expectations)
        new_color_rgb = (200, 100, 50)
        new_color_hex = "#c86432"
        mock_return = (new_color_rgb, new_color_hex)

        with patch('guibbon.widgets.color_picker_widget.askcolor', return_value=mock_return):
            widget.callback()

        # Check callback was called with the new color
        callback.assert_called_once_with((200, 100, 50))
        
        # Check widget internal state was updated
        self.assertEqual(widget.get_current_value(), (200, 100, 50))

    def test_callback_with_cancel(self) -> None:
        """Test callback when user cancels color picker dialog"""
        callback = Mock()
        initial_color = (50, 50, 50)
        widget = ColorPickerWidget(
            tk_frame=self.frame,
            name="Test",
            on_change=callback,
            initial_color_rgb=initial_color,
        )

        # Mock askcolor to return None (user cancelled)
        mock_return = (None, None)

        with patch('guibbon.widgets.color_picker_widget.askcolor', return_value=mock_return):
            widget.callback()

        # Callback should still be called (with original color)
        callback.assert_called_once()
        
        # Color should remain unchanged
        self.assertEqual(widget.get_current_value(), initial_color)
        self.assertEqual(widget.canvas["bg"], "#323232")

    def test_canvas_background_updates(self) -> None:
        """Test that canvas background color updates correctly"""
        callback = Mock()
        widget = ColorPickerWidget(
            tk_frame=self.frame,
            name="Test",
            on_change=callback,
            initial_color_rgb=(0, 0, 0),
        )

        # Check initial background
        self.assertEqual(widget.canvas["bg"], "#000000")

        # Simulate color change
        new_color_rgb = (255, 128, 64)
        new_color_hex = "#ff8040"
        mock_return = (new_color_rgb, new_color_hex)

        with patch('guibbon.widgets.color_picker_widget.askcolor', return_value=mock_return):
            widget.callback()

        # Check background was updated
        self.assertEqual(widget.canvas["bg"], "#ff8040")

    def test_widget_name_displayed(self) -> None:
        """Test that widget displays the correct name"""
        callback = Mock()
        widget_name = "My Color Picker"
        
        widget = ColorPickerWidget(
            tk_frame=self.frame,
            name=widget_name,
            on_change=callback,
        )

        # Find the label widget in the frame
        label_found = False
        for child in widget.frame.winfo_children():
            if isinstance(child, tk.Label):
                if child.cget("text") == widget_name:
                    label_found = True
                    break
        
        self.assertTrue(label_found, f"Label with text '{widget_name}' not found")

    def test_button_text(self) -> None:
        """Test that button has correct text"""
        callback = Mock()
        widget = ColorPickerWidget(
            tk_frame=self.frame,
            name="Test",
            on_change=callback,
        )

        self.assertEqual(widget.button.cget("text"), "Edit")

    def test_invalid_initial_color_negative(self) -> None:
        """Test that invalid initial color with negative value raises assertion"""
        callback = Mock()
        
        with self.assertRaises(AssertionError):
            ColorPickerWidget(
                tk_frame=self.frame,
                name="Test",
                on_change=callback,
                initial_color_rgb=(-1, 0, 0),  # type: ignore
            )

    def test_invalid_initial_color_too_large(self) -> None:
        """Test that invalid initial color > 255 raises assertion"""
        callback = Mock()
        
        with self.assertRaises(AssertionError):
            ColorPickerWidget(
                tk_frame=self.frame,
                name="Test",
                on_change=callback,
                initial_color_rgb=(256, 0, 0),  # type: ignore
            )

    def test_invalid_initial_color_wrong_length(self) -> None:
        """Test that initial color with wrong length raises assertion"""
        callback = Mock()
        
        with self.assertRaises(AssertionError):
            ColorPickerWidget(
                tk_frame=self.frame,
                name="Test",
                on_change=callback,
                initial_color_rgb=(255, 255),  # type: ignore
            )

    def test_multiple_color_changes(self) -> None:
        """Test multiple sequential color changes"""
        callback = Mock()
        widget = ColorPickerWidget(
            tk_frame=self.frame,
            name="Test",
            on_change=callback,
            initial_color_rgb=(0, 0, 0),
        )

        # First color change
        color1 = (100, 50, 25)
        with patch('guibbon.widgets.color_picker_widget.askcolor', return_value=(color1, "#64321a")):
            widget.callback()

        self.assertEqual(widget.get_current_value(), color1)

        # Second color change
        color2 = (200, 150, 100)
        with patch('guibbon.widgets.color_picker_widget.askcolor', return_value=(color2, "#c89664")):
            widget.callback()

        self.assertEqual(widget.get_current_value(), color2)

        # Verify callback was called twice
        self.assertEqual(callback.call_count, 2)


if __name__ == "__main__":
    unittest.main()



# type: ignore

import sys
import cv2
import guibbon as gbn
import numpy as np

if "guibbon" in sys.modules:
    gbn.inject(cv2)


def on_trackbar(val):
    print("on_trackbar", val)


def on_button_click():
    print("on_button_click")


def on_mouse_event(event, x, y, flags, param):
    print(event, x, y, flags, param)


def on_radio_button(i, opt):
    print("on_radio_button", i, opt)


def on_check_buttons(checks):
    print("on_check_buttons", checks)


def on_check_button(check):
    print("on_check_button", check)


def on_color_pick(colors):
    print("on_color_pick", colors)


def on_point_click(event):
    print("on_point_click", event)


def on_point_drag(event):
    print("on_point_drag", event)


def on_point_release(event):
    print("on_point_release", event)


def on_drag_multislider(event):
    print("on_drag_multislider", event)


def on_release_multislider(event):
    print("on_release_multislider", event)


def demo_cv():
    img = cv2.imread("ressources/dog.jpg")

    k = 0
    winname = "Demo Guibbon"
    cv2.namedWindow(winname)
    cv2.createTrackbar("trackbar_name", winname, 0, 10, on_trackbar)
    cv2.setTrackbarMin("trackbar_name", winname, 2)
    cv2.setTrackbarMax("trackbar_name", winname, 12)
    cv2.setTrackbarPos("trackbar_name", winname, 6)
    cv2.setMouseCallback(winname, on_mouse_event)

    if "guibbon" in sys.modules:
        # type: ignore
        radio_buttons = gbn.create_radio_buttons(winname, "radio2", ["peche", "prune"], on_radio_button)
        # gbn.createButton("more fruit", lambda : radio_buttons.set_options_list(["oups", "ok", "voilà"]), winname)
        button = gbn.create_button(winname, "more fruit", lambda: radio_buttons.set_options_list(["oups", "ok", "voilà"]))
        gbn.create_button(winname, "hide", lambda: button.set_visible(False))
        gbn.create_button(winname, "show", lambda: gbn.get_button_instance(winname, "more fruit").set_visible(True))

        gbn.create_check_button_list(winname, "check multi", ["roue", "volant"], on_check_buttons, [False, True])
        gbn.create_check_button(winname, "check single", on_check_button, initial_value=True)
        gbn.create_color_picker(winname, "Color picker", on_color_pick, (255, 0, 0))
        gbn.createInteractivePoint(winname, (100, 100), "point", on_click=on_point_click, on_drag=on_point_drag, on_release=on_point_release)

        multislider_values = "abcdefghijklmnopqrstuvwxyz"
        multislider = gbn.create_multislider(winname, "multi slider", multislider_values, initial_indexes=[0, 15], on_drag=on_drag_multislider, on_release=on_release_multislider)
        gbn.create_button(winname, "reset multi slider", lambda: multislider.set_positions(np.linspace(0, len(multislider_values)-1, num=len(multislider.get_positions()), dtype=int).tolist()))
        gbn.create_button(winname, "add cursor", lambda: multislider.add_cursor(len(multislider.get_positions())))
        gbn.create_button(winname, "remove cursor", lambda: multislider.remove_cursor())


        gbn.create_color_space_widget(winname, "color space", initial_color_space=[[255, 0, 0], [0, 255, 0], [0, 0, 255]], on_drag=lambda x:print("color space", x))

    cv2.namedWindow("ok")

    while gbn.Guibbon.is_instance(winname):  # cv2.getWindowProperty(title, cv2.WND_PROP_VISIBLE) > 0.5:
        k += 1
        cv2.imshow(winname, img, mode=gbn.MODE.P100)
        # gbn.imshow("ok", img * np.uint8(k))
        print(cv2.waitKeyEx(0), end=", ")


def print_hello():
    print("hello")


if __name__ == "__main__":
    demo_cv()



import unittest
import tkinter as tk

from guibbon.widgets.color_space_widget import ColorSpaceWidget, ColorSpace


# Shared Tk root for all tests to avoid tkinter instability
_tk_root = None


def setUpModule():
    """Create a single shared Tk root for all tests in this module"""
    global _tk_root
    _tk_root = tk.Tk()
    _tk_root.withdraw()  # Hide the window


def tearDownModule():
    """Destroy the shared Tk root after all tests"""
    global _tk_root
    if _tk_root:
        try:
            _tk_root.destroy()
        except tk.TclError:
            pass
        _tk_root = None


class TestColorSpaceWidget(unittest.TestCase):
    """Test suite for ColorSpaceWidget"""

    def setUp(self) -> None:
        """Set up test fixtures"""
        # Use a Frame as child of the shared root instead of creating new Tk()
        assert _tk_root is not None
        self.frame = tk.Frame(_tk_root, width=400, height=300)
        self.frame.pack(expand=True, fill='both')
        # Update to ensure frame is rendered and has dimensions
        _tk_root.update_idletasks()
        _tk_root.update()

    def tearDown(self) -> None:
        """Clean up after tests"""
        try:
            self.frame.destroy()
        except tk.TclError:
            pass

    def test_create_color_space_widget(self) -> None:
        """Test basic ColorSpaceWidget creation"""
        initial_color_space: ColorSpace = [(100, 100, 55), (155, 155, 200)]
        
        widget = ColorSpaceWidget(
            tk_frame=self.frame,
            color_space_name="Test Color Space",
            initial_color_space=initial_color_space,
        )
        
        # Verify widget was created
        self.assertIsNotNone(widget)
        self.assertEqual(widget.M, 2)
        self.assertEqual(len(widget.multi_slider_overlays), 3)  # RGB channels
        
        # Verify color space was properly stored
        self.assertEqual(len(widget.color_space), 2)
        self.assertEqual(widget.color_space[0], (100, 100, 55))
        self.assertEqual(widget.color_space[1], (155, 155, 200))

    def test_get_color_space(self) -> None:
        """Test getting the color space returns a copy"""
        initial_color_space: ColorSpace = [(50, 60, 70), (205, 195, 185)]
        
        widget = ColorSpaceWidget(
            tk_frame=self.frame,
            color_space_name="Test",
            initial_color_space=initial_color_space,
        )
        
        color_space = widget.get_color_space()
        
        # Verify we get a copy (not the same list)
        self.assertIsNot(color_space, widget.color_space)
        self.assertEqual(color_space, [(50, 60, 70), (205, 195, 185)])
        
        # Modifying returned list shouldn't affect widget
        color_space.append((0, 0, 0))
        self.assertEqual(len(widget.color_space), 2)

    def test_callbacks_none(self) -> None:
        """Test widget works with no callbacks"""
        initial_color_space: ColorSpace = [(128, 128, 127), (127, 127, 128)]
        
        widget = ColorSpaceWidget(
            tk_frame=self.frame,
            color_space_name="Test",
            initial_color_space=initial_color_space,
            on_drag=None,
            on_release=None,
        )
        
        # Should not raise any exceptions
        widget.on_drag_callback(0, [(100, 100)])
        widget.on_release_callback([(100, 100)])

    def test_on_drag_callback(self) -> None:
        """Test on_drag callback is called correctly"""
        initial_color_space: ColorSpace = [(100, 100, 100), (155, 155, 155)]
        
        callback_data = []
        def on_drag(color_space: ColorSpace) -> None:
            callback_data.append(("drag", color_space))
        
        widget = ColorSpaceWidget(
            tk_frame=self.frame,
            color_space_name="Test",
            initial_color_space=initial_color_space,
            on_drag=on_drag,
        )
        
        # Simulate drag callback
        widget.on_drag_callback(0, [(50, 50)])
        
        # Verify callback was called
        self.assertEqual(len(callback_data), 1)
        self.assertEqual(callback_data[0][0], "drag")
        self.assertIsInstance(callback_data[0][1], list)

    def test_on_release_callback(self) -> None:
        """Test on_release callback is called correctly"""
        initial_color_space: ColorSpace = [(100, 100, 100), (155, 155, 155)]
        
        callback_data = []
        def on_release(color_space: ColorSpace) -> None:
            callback_data.append(("release", color_space))
        
        widget = ColorSpaceWidget(
            tk_frame=self.frame,
            color_space_name="Test",
            initial_color_space=initial_color_space,
            on_release=on_release,
        )
        
        # Simulate release callback
        widget.on_release_callback([(50, 50)])
        
        # Verify callback was called
        self.assertEqual(len(callback_data), 1)
        self.assertEqual(callback_data[0][0], "release")
        self.assertIsInstance(callback_data[0][1], list)

    def test_update_color_space(self) -> None:
        """Test update_color_space updates the internal color space"""
        initial_color_space: ColorSpace = [(100, 50, 25), (155, 205, 230)]
        
        widget = ColorSpaceWidget(
            tk_frame=self.frame,
            color_space_name="Test",
            initial_color_space=initial_color_space,
        )
        
        # Call update_color_space
        widget.update_color_space()
        
        # Verify color_space is updated
        self.assertIsNotNone(widget.color_space)
        self.assertIsInstance(widget.color_space, list)

    def test_widget_color_parameter(self) -> None:
        """Test widget_color parameter is accepted"""
        initial_color_space: ColorSpace = [(128, 128, 128), (127, 127, 127)]
        
        widget = ColorSpaceWidget(
            tk_frame=self.frame,
            color_space_name="Test",
            initial_color_space=initial_color_space,
            widget_color="#FFFFFF",
        )
        
        self.assertIsNotNone(widget)


class TestColorSpaceConversion(unittest.TestCase):
    """Test suite for ColorSpaceWidget static conversion methods"""

    def test_rgb_segments_to_color_space_simple(self) -> None:
        """Test conversion from RGB segments to color space"""
        red_segments = [100]
        green_segments = [150]
        blue_segments = [200]
        
        color_space = ColorSpaceWidget.rgb_segments_to_color_space(
            red_segments, green_segments, blue_segments
        )
        
        # Expected: [(100, 150, 200), (155, 105, 55)]
        # First segment: (100, 150, 200) - (0, 0, 0)
        # Second segment: (255, 255, 255) - (100, 150, 200)
        self.assertEqual(len(color_space), 2)
        self.assertEqual(color_space[0], (100, 150, 200))
        self.assertEqual(color_space[1], (155, 105, 55))

    def test_rgb_segments_to_color_space_multiple(self) -> None:
        """Test conversion with multiple segments"""
        red_segments = [50, 100]
        green_segments = [60, 120]
        blue_segments = [70, 140]
        
        color_space = ColorSpaceWidget.rgb_segments_to_color_space(
            red_segments, green_segments, blue_segments
        )
        
        # Expected deltas:
        # Segment 0: (50, 60, 70) - (0, 0, 0) = (50, 60, 70)
        # Segment 1: (100, 120, 140) - (50, 60, 70) = (50, 60, 70)
        # Segment 2: (255, 255, 255) - (100, 120, 140) = (155, 135, 115)
        self.assertEqual(len(color_space), 3)
        self.assertEqual(color_space[0], (50, 60, 70))
        self.assertEqual(color_space[1], (50, 60, 70))
        self.assertEqual(color_space[2], (155, 135, 115))

    def test_color_space_to_rgb_segments_simple(self) -> None:
        """Test conversion from color space to RGB segments"""
        color_space: ColorSpace = [(100, 150, 200), (155, 105, 55)]
        
        red_segs, green_segs, blue_segs = ColorSpaceWidget.color_space_to_rgb_segments(
            color_space
        )
        
        # Expected cumulative sums (excluding final 255):
        # After (100, 150, 200): RGB = (100, 150, 200)
        # After (155, 105, 55): RGB = (255, 255, 255) - excluded
        self.assertEqual(red_segs, [100])
        self.assertEqual(green_segs, [150])
        self.assertEqual(blue_segs, [200])

    def test_color_space_to_rgb_segments_multiple(self) -> None:
        """Test conversion with multiple segments"""
        color_space: ColorSpace = [(50, 60, 70), (50, 60, 70), (155, 135, 115)]
        
        red_segs, green_segs, blue_segs = ColorSpaceWidget.color_space_to_rgb_segments(
            color_space
        )
        
        # Expected cumulative sums (excluding final 255):
        # After (50, 60, 70): RGB = (50, 60, 70)
        # After (50, 60, 70): RGB = (100, 120, 140)
        # After (155, 135, 115): RGB = (255, 255, 255) - excluded
        self.assertEqual(red_segs, [50, 100])
        self.assertEqual(green_segs, [60, 120])
        self.assertEqual(blue_segs, [70, 140])

    def test_conversion_roundtrip(self) -> None:
        """Test that conversion roundtrip preserves data"""
        # Start with RGB segments
        red_segments = [80, 160]
        green_segments = [90, 180]
        blue_segments = [100, 200]
        
        # Convert to color space
        color_space = ColorSpaceWidget.rgb_segments_to_color_space(
            red_segments, green_segments, blue_segments
        )
        
        # Convert back to RGB segments
        red_back, green_back, blue_back = ColorSpaceWidget.color_space_to_rgb_segments(
            color_space
        )
        
        # Should match original (sorted)
        self.assertEqual(red_back, sorted(red_segments))
        self.assertEqual(green_back, sorted(green_segments))
        self.assertEqual(blue_back, sorted(blue_segments))

    def test_rgb_segments_sorted_automatically(self) -> None:
        """Test that RGB segments are automatically sorted"""
        # Unsorted segments
        red_segments = [100, 50]
        green_segments = [150, 75]
        blue_segments = [200, 100]
        
        color_space = ColorSpaceWidget.rgb_segments_to_color_space(
            red_segments, green_segments, blue_segments
        )
        
        # Convert back - should be sorted
        red_back, green_back, blue_back = ColorSpaceWidget.color_space_to_rgb_segments(
            color_space
        )
        
        self.assertEqual(red_back, [50, 100])
        self.assertEqual(green_back, [75, 150])
        self.assertEqual(blue_back, [100, 200])

    def test_empty_segments(self) -> None:
        """Test conversion with empty segments (only 255)"""
        red_segments: list[int] = []
        green_segments: list[int] = []
        blue_segments: list[int] = []
        
        color_space = ColorSpaceWidget.rgb_segments_to_color_space(
            red_segments, green_segments, blue_segments
        )
        
        # Should have one segment: (255, 255, 255) - (0, 0, 0)
        self.assertEqual(len(color_space), 1)
        self.assertEqual(color_space[0], (255, 255, 255))

    def test_single_color_full_intensity(self) -> None:
        """Test color space with single full intensity color"""
        color_space: ColorSpace = [(255, 255, 255)]
        
        red_segs, green_segs, blue_segs = ColorSpaceWidget.color_space_to_rgb_segments(
            color_space
        )
        
        # All segments should be empty (final 255 is removed)
        self.assertEqual(red_segs, [])
        self.assertEqual(green_segs, [])
        self.assertEqual(blue_segs, [])


if __name__ == "__main__":
    unittest.main()



import unittest

from guibbon import mouse_pan
from guibbon import transform_matrix as tmat
import tkinter as tk

EPS = 1e-15


class TestMousePan(unittest.TestCase):
    def reset_counters(self):
        self.on_click_count = 0
        self.on_drag_count = 0
        self.on_release_count = 0

    def on_click(self, p0_xy):
        self.on_click_count += 1
        self.p0_xy = p0_xy

    def on_drag(self, p0_xy, p1_xy):
        self.on_drag_count += 1
        self.p0_xy = p0_xy
        self.p1_xy = p1_xy

    def on_release(self, p0_xy, p1_xy):
        self.on_release_count += 1
        self.p0_xy = p0_xy
        self.p1_xy = p1_xy

    def setUp(self):
        self.on_click_count = 0
        self.on_drag_count = 0
        self.on_release_count = 0
        self.p0_xy = (-1, -1)
        self.p1_xy = (-1, -1)
        self.right_button_num = 3
        self.mp = mouse_pan.MousePan(button_num=self.right_button_num, on_click=self.on_click, on_drag=self.on_drag, on_release=self.on_release)

    def test_MousePan_callback_triggering(self):
        self.assertFalse(self.mp.is_down, "MousePan.is_down must be initialized to False")
        event: tk.Event = tk.Event()  # type: ignore
        event.x = 1
        event.y = 1
        event.num = self.right_button_num

        self.reset_counters()
        event.type = tk.EventType.Motion
        self.mp.on_tk_event(event, tmat.identity_matrix())
        self.assertFalse(self.mp.is_down, 'is_down should remain the same on "Motion" event')
        self.assertEqual(self.on_click_count, 0, 'on_click callback must NOT be triggered with EventType "Motion" when is_down is False')
        self.assertEqual(self.on_drag_count, 0, 'on_drag callback must NOT be triggered with EventType "Motion" when is_down is False')
        self.assertEqual(self.on_release_count, 0, 'on_release callback must NOT be triggered with EventType "Motion" when is_down is False')

        self.reset_counters()
        event.type = tk.EventType.ButtonRelease
        self.mp.on_tk_event(event, tmat.identity_matrix())
        self.assertFalse(self.mp.is_down, '"ButtonRelease" event should have no effect when NOT is_down')
        self.assertEqual(self.on_click_count, 0, 'on_click callback must NOT be triggered with EventType "ButtonRelease" when is_down is False')
        self.assertEqual(self.on_drag_count, 0, 'on_drag callback must NOT be triggered with EventType "ButtonRelease" when is_down is False')
        self.assertEqual(self.on_release_count, 0, 'on_release callback must NOT be triggered with EventType "ButtonRelease" when is_down is False')

        self.reset_counters()
        event.type = tk.EventType.ButtonPress
        self.mp.on_tk_event(event, tmat.identity_matrix())
        self.assertTrue(self.mp.is_down, '"ButtonPress" event should should set "isdown" to True')
        self.assertEqual(self.on_click_count, 1, 'on_click callback must be triggered with EventType "ButtonPress')
        self.assertEqual(self.on_drag_count, 0, 'on_drag callback must NOT be triggered with EventType "ButtonPress')
        self.assertEqual(self.on_release_count, 0, 'on_release callback must NOT be triggered with EventType "ButtonPress')

        self.reset_counters()
        event.type = tk.EventType.Motion
        self.mp.on_tk_event(event, tmat.identity_matrix())
        self.assertTrue(self.mp.is_down, 'is_down should remain the same on "Motion" event')
        self.assertEqual(self.on_click_count, 0, 'on_click callback must NOT be triggered with EventType "Motion"')
        self.assertEqual(self.on_drag_count, 1, 'on_drag callback must be triggered with EventType "Motion')
        self.assertEqual(self.on_release_count, 0, 'on_release callback must NOT be triggered with EventType "Motion')

        self.reset_counters()
        event.type = tk.EventType.ButtonRelease
        self.mp.on_tk_event(event, tmat.identity_matrix())
        self.assertFalse(self.mp.is_down, 'is_down should be False after "ButtonRelease"')
        self.assertEqual(self.on_click_count, 0, 'on_click callback must NOT be triggered with EventType "ButtonRelease" when is_down is False')
        self.assertEqual(self.on_drag_count, 0, 'on_drag callback must NOT be triggered with EventType "ButtonRelease" when is_down is False')
        self.assertEqual(self.on_release_count, 1, 'on_release callback must be triggered with EventType "ButtonRelease" when is_down is False')

    def test_MousePan_values(self):
        def tuple_is_same(tup1, tup2, eps=EPS):
            return tuple(abs(v1 - v2) < eps for v1, v2 in zip(tup1, tup2))

        event: tk.Event = tk.Event()  # type: ignore
        x0, y0 = 11, 22
        x1, y1 = 111, 222
        event.x, event.y = x0, y0
        event.num = self.right_button_num

        sx, sy = (3, 4)
        scale_mat = tmat.scale_matrix((sx, sy))

        event.type = tk.EventType.ButtonPress
        self.mp.on_tk_event(event, scale_mat)
        self.assertTupleEqual((x0, y0), (event.x, event.y), "MousePan must NOT modify the event value passed as argument")
        self.assertTupleEqual(tuple_is_same((sx * x0, sy * y0), self.mp.p0_xy), (True, True), "MousePan must correctly convert canvas coordinates to image coordinates")
        self.assertTupleEqual(tuple_is_same((sx * x0, sy * y0), self.mp.p1_xy), (True, True), "MousePan must correctly convert canvas coordinates to image coordinates")

        event.type = tk.EventType.Motion
        event.x, event.y = x1, y1
        self.mp.on_tk_event(event, scale_mat)
        self.assertTupleEqual((x1, y1), (event.x, event.y), "MousePan must NOT modify the event value passed as argument")
        self.assertTupleEqual(tuple_is_same((sx * x0, sy * y0), self.mp.p0_xy), (True, True), "MousePan must correctly convert canvas coordinates to image coordinates")
        self.assertTupleEqual(tuple_is_same((sx * x1, sy * y1), self.mp.p1_xy), (True, True), "MousePan must correctly convert canvas coordinates to image coordinates")

        event.type = tk.EventType.ButtonRelease
        self.mp.on_tk_event(event, scale_mat)
        self.assertTupleEqual((x1, y1), (event.x, event.y), "MousePan must NOT modify the event value passed as argument")
        self.assertTupleEqual(tuple_is_same((sx * x0, sy * y0), self.mp.p0_xy), (True, True), "MousePan must correctly convert canvas coordinates to image coordinates")
        self.assertTupleEqual(tuple_is_same((sx * x1, sy * y1), self.mp.p1_xy), (True, True), "MousePan must correctly convert canvas coordinates to image coordinates")


if __name__ == "__main__":
    unittest.main()





import tkinter as tk
from typing import Sequence, Optional

from guibbon.transform_matrix import TransformMatrix
from guibbon.typedef import Point2D, Point2DList, CallbackPoint, CallbackPolygon, CallbackRect, InteractivePolygon
from .base import State, Point, Magnets


class Polygon(InteractivePolygon):
    colors = {
        State.NORMAL: "#%02x%02x%02x" % (0, 255, 0),
        State.HOVERED: "#%02x%02x%02x" % (127, 255, 127),
        State.DRAGGED: "#%02x%02x%02x" % (255, 255, 0),
    }

    def __init__(
        self,
        canvas: tk.Canvas,
        point_xy_list: Point2DList,
        label: str = "",
        on_click: CallbackPolygon = None,
        on_drag: CallbackPolygon = None,
        on_release: CallbackPolygon = None,
        img2can_matrix: Optional[TransformMatrix] = None,
        magnets: Optional[Magnets] = None,
    ):
        self.canvas = canvas
        self.label = label
        self.visible: bool = True
        self.state: State = State.NORMAL

        self.on_click = on_click
        self.on_drag = on_drag
        self.on_release = on_release
        N = len(point_xy_list)

        # create on_drag lambdas to pass point index and fix value of k
        def make_on_drag(self, k: int):
            return lambda event: self._on_drag(k, event)

        on_drag_lambdas: Sequence[CallbackPoint] = [make_on_drag(self, k) for k in range(N)]

        self.ipoints = []
        for k, point_xy in enumerate(point_xy_list):
            # subscribe to on_click only if needed.
            # subscribe to on_drag in any cases (to update points coordinates). Use lambda to pass point position.
            # subscribe to on_release only if needed.
            ipoint = Point(
                canvas,
                point_xy,
                label="",
                on_click=None if on_click is None else self._on_click,
                on_drag=on_drag_lambdas[k],
                on_release=None if on_release is None else self._on_release,
                img2can_matrix=img2can_matrix,
                magnets=magnets,
            )
            self.ipoints.append(ipoint)

        self.lines = self._create_lines()
        self.update()

    def _create_lines(self):
        lines = []
        N = len(self.ipoints)
        for i in range(N):
            i1 = i
            i2 = (i + 1) % N
            line_id = self.canvas.create_line(-1, -1, -1, -1, fill=Polygon.colors[self.state], width=5)
            lines.append((i1, i2, line_id))
        return lines

    def _update_lines(self):
        # draw lines
        item_state = "normal" if self.visible else "hidden"
        point_xy_list = [ipoint.get_can_point_xy() for ipoint in self.ipoints]
        point_xy_list = [(int(round(x)), int(round(y))) for x, y in point_xy_list]
        for i1, i2, line_id in self.lines:
            x1, y1 = point_xy_list[i1]
            x2, y2 = point_xy_list[i2]
            self.canvas.coords(line_id, x1, y1, x2, y2)
            self.canvas.itemconfig(line_id, state=item_state)
            self.canvas.tag_raise(line_id)

    def _update_points(self):
        for ipoint in self.ipoints:
            ipoint.update()

    def update(self):
        self._update_lines()
        self._update_points()

    def _on_click(self, event):
        if self.on_click is not None:
            point_xy_list = [ipoint.get_img_point_xy() for ipoint in self.ipoints]
            self.on_click(event, point_xy_list)

    def _on_drag(self, i, event):
        try:
            self._update_lines()
            if self.on_drag is not None:
                point_xy_list = [ipoint.get_img_point_xy() for ipoint in self.ipoints]
                self.on_drag(event, point_xy_list)
        except Exception as e:
            print(f"ERROR: {self}: self._on_drag({event}) --->", e)
            raise e

    def _on_release(self, event):
        if self.on_release is not None:
            point_xy_list = [ipoint.get_img_point_xy() for ipoint in self.ipoints]
            self.on_release(event, point_xy_list)

    def set_img2can_matrix(self, img2can_matrix: TransformMatrix):
        for ipoint in self.ipoints:
            ipoint.set_img2can_matrix(img2can_matrix)

    def set_point_xy_list(self, point_xy_list: Point2DList):
        assert len(point_xy_list) == len(self.ipoints)
        for ipoint, point_xy in zip(self.ipoints, point_xy_list):
            ipoint.set_img_point_xy(point_xy)
            self.update()

    def set_visible(self, value: bool):
        if value == self.visible:
            return
        self.visible = value
        for ipoint in self.ipoints:
            ipoint.set_visible(value)


class Rectangle(Polygon):
    def __init__(
        self,
        canvas: tk.Canvas,
        point0_xy: Point2D,
        point1_xy: Point2D,
        label: str = "",
        on_click: CallbackRect = None,
        on_drag: CallbackRect = None,
        on_release: CallbackRect = None,
        img2can_matrix: Optional[TransformMatrix] = None,
        magnets: Optional[Magnets] = None,
    ):
        # wrap user callback to convert signature from CallbackRect to CallbackPolygon
        lambda0 = None if on_click is None else lambda event, point_list_xy: on_click(event, point_list_xy[0], point_list_xy[1])
        on_click_rect: CallbackPolygon = lambda0

        lambda1 = None if on_drag is None else lambda event, point_list_xy: on_drag(event, point_list_xy[0], point_list_xy[1])
        on_drag_rect: CallbackPolygon = lambda1

        lambda2 = None if on_release is None else lambda event, point_list_xy: on_release(event, point_list_xy[0], point_list_xy[1])
        on_release_rect: CallbackPolygon = lambda2

        # Create a Point2DList from the 2 Point2D
        point_list_xy = [point0_xy, point1_xy]

        # create an instance of Polygon using wrapped callbacks
        super().__init__(
            canvas,
            point_list_xy,
            label,
            on_click=None if on_click is None else on_click_rect,
            on_drag=None if on_drag is None else on_drag_rect,
            on_release=None if on_release is None else on_release_rect,
            img2can_matrix=img2can_matrix,
            magnets=magnets,
        )

    def _create_lines(self):
        return [(-1, -1, self.canvas.create_line(-1, -1, -1, -1, fill=Polygon.colors[self.state], width=5)) for i in range(4)]

    def _update_lines(self):
        item_state = "normal" if self.visible else "hidden"

        point_xy_list = [ipoint.get_can_point_xy() for ipoint in self.ipoints]
        point_xy_list = [(int(round(x)), int(round(y))) for x, y in point_xy_list]

        left = point_xy_list[0][0]
        top = point_xy_list[0][1]
        right = point_xy_list[1][0]
        bottom = point_xy_list[1][1]

        line_id = self.lines[0][2]
        self.canvas.coords(line_id, left, top, right, top)
        self.canvas.itemconfig(line_id, state=item_state)
        self.canvas.tag_raise(line_id)

        line_id = self.lines[1][2]
        self.canvas.coords(line_id, right, top, right, bottom)
        self.canvas.itemconfig(line_id, state=item_state)
        self.canvas.tag_raise(line_id)

        line_id = self.lines[2][2]
        self.canvas.coords(line_id, right, bottom, left, bottom)
        self.canvas.itemconfig(line_id, state=item_state)
        self.canvas.tag_raise(line_id)

        line_id = self.lines[3][2]
        self.canvas.coords(line_id, left, bottom, left, top)
        self.canvas.itemconfig(line_id, state=item_state)
        self.canvas.tag_raise(line_id)



import tkinter as tk


class Entry:
    def __init__(self, master, on_change=None, on_return=None):
        self.on_change = (lambda x: None) if on_change is None else on_change
        self.on_return = (lambda x: None) if on_return is None else on_return

        self.is_focus = False
        self.text_var = tk.StringVar()

        # Tcl doc about validatecommand: https://tcl-lang.org/man/tcl8.6/TkCmd/ttk_entry.htm#M34
        cnf = {
            "textvariable": self.text_var,
            "width": 8,
        }
        if on_change is not None:
            vcmd = master.register(self._on_change)
            cnf.update({
                "validate": "all",
                "validatecommand": (vcmd, "%P"),
            })

        self.entry: tk.Entry = tk.Entry(master=master, cnf=cnf)
        self.entry.bind('<Return>', self._on_return)
        self.entry.bind("<FocusOut>", self._on_focus_out)


    def pack(self, cnf=None, **kw):
        if cnf is None:
            cnf = {}

        self.entry.pack(cnf, **kw)

    def set(self, text: str):
        self.text_var.set(text)

    def get(self):
        return self.text_var.get()

    def _on_change(self, text: str) -> bool:
        self.on_change(text)
        self.is_focus = True
        return True

    def _on_return(self, event):
        # focus out by giving focus to the next item

        try:
            self.entry.tk_focusNext().focus()  # type: ignore
        except AttributeError:
            pass
        self.on_return(self.text_var.get())

    def _on_focus_out(self, event):
        # focus out by giving focus to the next item
        self.is_focus = False



import unittest

import numpy as np

from guibbon import transform_matrix as tmat


def maxdiff(arr1, arr2):
    return np.max(np.abs(np.array(arr1) - np.array(arr2)).flatten())


eps = 1e-15


class TestTransformMatrix(unittest.TestCase):
    def test_IsTransformMatrix(self):
        self.assertTrue(tmat.isTransformMatrix(np.array([[1, 0, 0], [0, 1, 0], [0, 0, 1]], dtype=np.float64)))
        self.assertTrue(tmat.isTransformMatrix(np.array([[1, 0, 0], [0, 1, 0], [0, 0, 1]], dtype=float)))
        self.assertFalse(tmat.isTransformMatrix(np.array([[1, 0, 0], [0, 1, 0], [0, 0, 1]], dtype=np.float32)))
        self.assertFalse(tmat.isTransformMatrix(np.array([[1, 0, 0], [0, 1, 0], [0, 0, 1]], dtype=np.int32)))
        self.assertFalse(tmat.isTransformMatrix(np.array([[1, 0, 0], [0, 1, 0], [0, 0, 1]], dtype=np.int64)))
        self.assertFalse(tmat.isTransformMatrix(np.array([[1, 0, 0], [0, 1, 0]], dtype=np.float64)))
        self.assertFalse(tmat.isTransformMatrix(np.array([[1, 0], [0, 1], [0, 0]], dtype=np.float64)))

    def test_identity_matrix(self):
        expected = np.array([[1, 0, 0], [0, 1, 0], [0, 0, 1]])

        mat1 = tmat.identity_matrix()
        mat2 = tmat.I()

        for mat in [mat1, mat2]:
            self.assertTrue(mat.dtype == float)
            self.assertTupleEqual(mat.shape, (3, 3))
            self.assertLess(maxdiff(mat, expected), eps)

    def test_translation_matrix(self):
        trans_list = [(0, 0), (1.1, 2.2), (-3, 19.2)]
        for tx, ty in trans_list:
            expected = np.array([[1, 0, tx], [0, 1, ty], [0, 0, 1]])

            mat1 = tmat.translation_matrix((tx, ty))
            mat2 = tmat.T((tx, ty))

            for mat in [mat1, mat2]:
                self.assertTrue(mat.dtype == float)
                self.assertTupleEqual(mat.shape, (3, 3))
                self.assertLess(maxdiff(mat, expected), eps)

        self.assertLess(maxdiff(tmat.T((0, 0)), tmat.I()), eps)

    def test_scale_matrix(self):
        scale_list = [(0, 0), (1, 1), (-1, -1), (1.0, 2.3)]
        for sx, sy in scale_list:
            expected = np.array([[sx, 0, 0], [0, sy, 0], [0, 0, 1]])

            mat1 = tmat.scale_matrix((sx, sy))
            mat2 = tmat.S((sx, sy))

            for mat in [mat1, mat2]:
                self.assertTrue(mat.dtype == float)
                self.assertTupleEqual(mat.shape, (3, 3))
                self.assertLess(maxdiff(mat, expected), eps)

        self.assertLess(maxdiff(tmat.S((1, 1)), tmat.I()), eps)

    def test_rotation_matrix(self):
        thetas_list = [0, -1, 1, 1.2]
        for theta in thetas_list:
            cs, sn = np.cos(theta), np.sin(theta)
            expected = np.array([[cs, -sn, 0], [sn, cs, 0], [0, 0, 1]])

            mat1 = tmat.rotation_matrix(theta)
            mat2 = tmat.R(theta)

            for mat in [mat1, mat2]:
                self.assertTrue(mat.dtype == float)
                self.assertTupleEqual(mat.shape, (3, 3))
                self.assertLess(maxdiff(mat, expected), eps)

        self.assertLess(maxdiff(tmat.R(0), tmat.I()), eps)
        self.assertLess(maxdiff(tmat.R(np.pi * 2), tmat.I()), eps)

    def test_apply(self):
        def assertIsPoint2D(point_xy):
            self.assertTrue(isinstance(point_xy, tuple), f"Point2D must be a tuple, got {type(point_xy)} instead")
            self.assertEqual(len(point_xy), 2, "Point2D must be of size 2")
            for val in point_xy:
                self.assertTrue(isinstance(val, int) or isinstance(val, float), "Point2D must contains numeric values (either int or float)")

        for point_xy in [(0, 0), (0.0, 0.0)]:
            point_res = tmat.apply(tmat.I(), point_xy)
            assertIsPoint2D(point_res)

        point_xy_list = [(0, 0), (1, 1), (-1, -1), (1.0, 2.3)]
        tx, ty = 10, -20.5
        sx, sy = 2, -2.3
        mat = tmat.T((tx, ty)) @ tmat.S((sx, sy))
        for x, y in point_xy_list:
            x_expected = x * sx + tx
            y_expected = y * sy + ty
            point_expected = x_expected, y_expected
            point_res = tmat.apply(mat, (x, y))
            assertIsPoint2D(point_res)
            self.assertLess(maxdiff(point_res, point_expected), eps)

        with self.assertRaises(TypeError):
            tmat.apply(np.array([1]), (0, 0))


if __name__ == "__main__":
    unittest.main()



import unittest
from guibbon.typedef import InteractivePoint, InteractivePolygon
from types import FunctionType


class TestPoint(unittest.TestCase):
    def test_interface_pattern(self):
        for interface in [InteractivePoint, InteractivePolygon]:
            with self.assertRaises(AttributeError, msg="Calling non existing functions must raise an AttributeError"):
                _ = interface.this_does_not_exist_and_should_raise  # type: ignore

            # makes sure the interface correctly prevent bad implementation
            class BadClass(interface):  # type: ignore
                pass

            with self.assertRaises(TypeError, msg="BadClass doesn't properly implement the interface. Its instanciation must raise a TypeError"):
                BadClass()

    def test_InteractivePoint_interface(self):
        func_list = [
            InteractivePoint.update,
            InteractivePoint.set_img_point_xy,
            InteractivePoint.set_visible,
        ]
        for func in func_list:
            self.assertIsInstance(func, FunctionType, msg=f"Attribute {func} is supposed to be a function")

    def test_InteractivePolygon_interface(self):
        func_list = [
            InteractivePolygon.update,
            InteractivePolygon.set_point_xy_list,
            InteractivePolygon.set_visible,
        ]
        for func in func_list:
            self.assertIsInstance(func, FunctionType, msg=f"Attribute {func} is supposed to be a function")


if __name__ == "__main__":
    unittest.main()



import dataclasses
import numpy as np
import cv2
import guibbon as gbn
from guibbon.typedef import Image_t
import threading
import tkinter as tk


def on_click_treeview(*args):
    print(args)


class DemoMVCAdvApp:
    @dataclasses.dataclass
    class Model:
        x: int = 0
        y: int = 0
        cross_xy = (0, 0)

    @dataclasses.dataclass
    class Result:
        img: Image_t = dataclasses.field(default_factory=lambda: np.zeros((1, 1, 3), dtype=np.uint8))

    def __init__(self, filename):
        img = cv2.imread(filename)
        if img is None:
            raise FileNotFoundError(f"Image file not found: {filename}")
        self.img: Image_t = img.astype(np.uint8)
        self.winname = "demo app"
        self.lock = threading.Lock()
        gbn.namedWindow(self.winname)
        gbn.createInteractivePoint(self.winname, (100, 100), "point", on_drag=self.on_drag, magnet_points=[(200, 220), (30, 30)])

        point_xy_list = [(300, 300), (300, 500), (500, 400), (400, 300)]
        gbn.createInteractivePolygon(self.winname, point_xy_list, "polygon", on_drag=self.on_drag_poly, magnet_points=[(100, 300), (170, 300)])

        point0_xy = (200, 200)
        point1_xy = (300, 250)
        gbn.createInteractiveRectangle(self.winname, point0_xy, point1_xy, "rectangle", on_drag=self.on_drag_rect, magnet_points=[(100, 300), (150, 300)])

        tree = gbn.TreeNode(name="root", value="r")
        tree.set(["branch0"], value="0")
        tree.set(["branch0", "branch00"], value="00")
        tree.set(["branch0", "branch01"], value="01")
        tree.set(["branch1", "branch10"], value="10")
        gbn.create_treeview(self.winname, "TreeView", tree=tree, on_click=on_click_treeview)

        self.model = DemoMVCAdvApp.Model()
        self.result = DemoMVCAdvApp.Result()
        self.is_update_needed: bool = True

    def update(self):
        h, w = self.img.shape[:2]
        with self.lock:
            x = self.model.x / w
            y = self.model.y / h
            cross_x, cross_y = self.model.cross_xy

        gain_b = max(0.5, min(2.0, 1.5 * x + 0.5))
        gain_r = max(0.5, min(2.0, 1.5 * y + 0.5))

        res = self.img.copy()
        res = res.astype(np.uint16)
        res[:, :, 0:1] = res[:, :, 0:1] * gain_b
        res[:, :, 2:3] = res[:, :, 2:3] * gain_r
        res[res > 255] = 255

        res[int(round(cross_y)), :, 2] = 255
        res[:, int(round(cross_x)), 2] = 255
        res = res.astype(np.uint8)
        self.result.img = res
        self.is_update_needed = False

    def show(self):
        while True:
            if self.is_update_needed:
                self.update()
                gbn.imshow(self.winname, self.result.img, mode=gbn.MODE.FIT)
                self.is_update_needed = False
            gbn.waitKeyEx(10)

    def on_drag(self, event: tk.Event) -> None:
        print(f"user.on_drag: ({event.x}, {event.y})")
        with self.lock:
            self.model.x = event.x
            self.model.y = event.y
            self.model.cross_xy = (event.x, event.y)
            # print(event.x, event.y)
            self.is_update_needed = True

    def on_drag_poly(self, event: tk.Event, point_xy_list: gbn.Point2DList) -> None:
        # print(point_xy_list)
        with self.lock:
            self.point_xy_list = point_xy_list + []
            self.model.cross_xy = (event.x, event.y)
            self.is_update_needed = True

    def on_drag_rect(self, event: tk.Event, point0_xy: gbn.Point2D, point1_xy: gbn.Point2D) -> None:
        print(point0_xy, point1_xy)
        with self.lock:
            self.rect_p0_xy = point0_xy
            self.rect_p1_xy = point1_xy
            self.model.cross_xy = (event.x, event.y)
            self.is_update_needed = True


if __name__ == "__main__":
    mc = DemoMVCAdvApp("ressources/dog.jpg")
    mc.show()



import unittest
import tkinter as tk
from guibbon.widgets.multi_slider_widget import MultiSliderOverlay


# Shared Tk root for all tests to avoid tkinter instability
_tk_root = None


def setUpModule():
    """Create a single shared Tk root for all tests in this module"""
    global _tk_root
    _tk_root = tk.Tk()
    _tk_root.withdraw()  # Hide the window


def tearDownModule():
    """Destroy the shared Tk root after all tests"""
    global _tk_root
    if _tk_root:
        try:
            _tk_root.destroy()
        except tk.TclError:
            pass
        _tk_root = None


class TestMultiSliderOverlay(unittest.TestCase):
    def setUp(self):
        # Use the shared root and create a canvas as a child
        self.canvas = tk.Canvas(_tk_root, width=200, height=30)
        self.callback_count = 0
        self.last_positions_values = None

    def tearDown(self):
        try:
            self.canvas.destroy()
        except tk.TclError:
            pass

    def callback(self, positions_values):
        self.callback_count += 1
        self.last_positions_values = positions_values

    def test_create_overlay(self):
        """Test basic creation of MultiSliderOverlay"""
        values = list(range(10))
        initial_positions = [2, 5]
        
        overlay = MultiSliderOverlay(
            canvas=self.canvas,
            values=values,
            initial_positions=initial_positions,
            on_drag=self.callback,
            on_release=self.callback
        )
        
        self.assertEqual(len(overlay.cursors), 2)
        self.assertEqual(overlay.get_positions(), initial_positions)

    def test_canvas_to_slider_conversion(self):
        """Test coordinate conversion from canvas to slider"""
        values = list(range(10))
        initial_positions = [0]
        
        overlay = MultiSliderOverlay(
            canvas=self.canvas,
            values=values,
            initial_positions=initial_positions
        )
        
        # Update canvas to get proper dimensions
        self.canvas.update_idletasks()
        
        # Test conversion (basic sanity check)
        slider_point = overlay.canvas2slider((100, 15))
        self.assertIsInstance(slider_point, tuple)
        self.assertEqual(len(slider_point), 2)
        
        # x should be within valid range
        x, y = slider_point
        self.assertGreaterEqual(x, 0)
        self.assertLessEqual(x, len(values) - 1)

    def test_slider_to_canvas_conversion(self):
        """Test coordinate conversion from slider to canvas"""
        values = list(range(10))
        initial_positions = [0]
        
        overlay = MultiSliderOverlay(
            canvas=self.canvas,
            values=values,
            initial_positions=initial_positions
        )
        
        # Update canvas to get proper dimensions
        self.canvas.update_idletasks()
        
        # Test conversion
        canvas_point = overlay.slider2canvas((5, 0))
        self.assertIsInstance(canvas_point, tuple)
        self.assertEqual(len(canvas_point), 2)
        
        # Coordinates should be within canvas bounds
        x, y = canvas_point
        self.assertGreaterEqual(x, 0)
        self.assertLessEqual(x, self.canvas.winfo_width())


if __name__ == "__main__":
    unittest.main()



import dataclasses
import sys
import tkinter
import unittest

from guibbon import interactive_overlays
from guibbon import transform_matrix as tmat
from guibbon.typedef import InteractivePoint

eps = sys.float_info.epsilon


@dataclasses.dataclass
class Event:
    x: int = 0
    y: int = 0


class EventName:
    CLICK = "<Button-1>"
    DRAG = "<B1-Motion>"
    RELEASE = "<ButtonRelease-1>"
    ENTER = "<Enter>"
    LEAVE = "<Leave>"


class TestMagnets(unittest.TestCase):
    def foo(self, x_img, y_img):
        return self.val, self.val

    def setUp(self) -> None:
        self.canvas = tkinter.Canvas()
        self.val = 0
        self.points_img = [(10.1 * k, 10.1 * k) for k in range(4)]
        self.img2can_matrix = tmat.T((123, 456)) @ tmat.S((100, 200))

    def test_magnets_creation(self):
        magnets = interactive_overlays.Magnets(self.canvas, self.points_img)
        self.assertListEqual(tmat.I().tolist(), magnets.img2can_matrix.tolist(), "Default img2can_matrix must be indentity")
        magnets.set_img2can_matrix(self.img2can_matrix)
        self.assertListEqual(self.img2can_matrix.tolist(), magnets.img2can_matrix.tolist(), "img2can_matrix must be settable")
        self.assertNotEqual(id(self.img2can_matrix), id(magnets.img2can_matrix), "set_img2can_matrix must deeply copy the matrix")

        # giving empty Point2DList should not raise any exception
        interactive_overlays.Magnets(self.canvas, [])

        # default threshold is bigger than 0
        self.assertTrue(magnets.dist_threshold > 0)

    def test_snap_to_nearest_magnet(self):
        empty_magnet = interactive_overlays.Magnets(self.canvas, [])
        for point_xy in [(0, 0), (10, 20), (300, 400), (-10.2, -12.2)]:
            self.assertTupleEqual(point_xy, empty_magnet.snap_to_nearest_magnet(point_xy), "empty magnet should not snap")

        magnets = interactive_overlays.Magnets(self.canvas, self.points_img)
        x, y = self.points_img[-1]
        for thresh in [10, 50]:
            magnets.dist_threshold = thresh
            self.assertTupleEqual((x, y), magnets.snap_to_nearest_magnet((x, y)))
            self.assertTupleEqual((x, y), magnets.snap_to_nearest_magnet((x + thresh - 1, y)))
            self.assertTupleEqual((x + thresh + 1, y), magnets.snap_to_nearest_magnet((x + thresh + 1, y)))

        magnets.dist_threshold = 50
        self.assertTupleEqual((10.1, 10.1), magnets.snap_to_nearest_magnet((14, 14)))
        self.assertTupleEqual((20.2, 20.2), magnets.snap_to_nearest_magnet((16, 16)))

    def test_update(self):
        magnets = interactive_overlays.Magnets(self.canvas, self.points_img)
        magnets.set_img2can_matrix(self.img2can_matrix)

        self.assertEqual(len(self.points_img), len(magnets.circle_id_list), "Each points must have its circle on the canvas")

        magnets.update()
        for point_img_xy, circle_id in zip(self.points_img, magnets.circle_id_list):
            x_can_expected, y_can_expected = tmat.apply(self.img2can_matrix, point_img_xy)
            x1, y1, x2, y2 = self.canvas.coords(circle_id)
            x = round((x1 + x2) / 2)
            y = round((y1 + y2) / 2)
            self.assertLess(abs(x - x_can_expected), 0.1, "Coordinates of magnet points must be accurately placed on the canvas")
            self.assertLess(abs(y - y_can_expected), 0.1, "Coordinates of magnet points must be accurately placed on the canvas")


class TestPoint(unittest.TestCase):
    def setUp(self) -> None:
        self.event_count = 0
        self.drag_count = 0
        self.release_count = 0
        self.point_xy = (0, 0)

        self.canvas = tkinter.Canvas()
        self.magnets = interactive_overlays.Magnets(self.canvas, [(1000.0, 1000.0)], (lambda x, y: (2 * x, 2 * y)))

    def tearDown(self) -> None:
        pass

    def on_event(self, event):
        self.event_count += 1
        self.point_xy = (event.x, event.y)

    def on_event_error(self, event):
        raise

    def test_interface(self):
        self.pt = interactive_overlays.Point(canvas=self.canvas, point_xy=self.point_xy)
        self.assertIsInstance(self.pt, InteractivePoint)

    def test_callback(self):
        self.pt = interactive_overlays.Point(
            canvas=self.canvas, point_xy=self.point_xy, label="ok", on_click=self.on_event, on_drag=self.on_event, on_release=self.on_event
        )

        binds = self.canvas.tag_bind(self.pt.circle_id)
        self.assertIn(EventName.CLICK, binds, f"{EventName.CLICK} tag not binded")
        self.assertIn(EventName.DRAG, binds, f"{EventName.DRAG} tag not binded")
        self.assertIn(EventName.RELEASE, binds, f"{EventName.RELEASE} tag not binded")
        self.assertIn(EventName.ENTER, binds, f"{EventName.ENTER} tag not binded")
        self.assertIn(EventName.LEAVE, binds, f"{EventName.LEAVE} tag not binded")

        self.assertEqual(0, self.event_count)
        self.assertEqual((0, 0), self.point_xy)

        pt_point_bck = self.pt.point_xy
        event = Event(x=11, y=12)
        self.pt._on_click(event)
        self.assertEqual(1, self.event_count, "Event must be triggered")
        self.assertEqual((11, 12), self.point_xy, "Event coordinate must be passed to callback")
        self.assertEqual(pt_point_bck, self.pt.point_xy, "point._on_click() must not set point_xy coordinates")

        event = Event(x=21, y=22)
        self.pt._on_drag(event)
        self.assertEqual(2, self.event_count, "Event must be triggered")
        self.assertEqual((21, 22), self.point_xy, "Event coordinate must be passed to callback")
        self.assertEqual((21, 22), self.pt.point_xy, "point._on_drag() must correctly set point_xy coordinates")

        pt_point_bck = self.pt.point_xy
        event = Event(x=31, y=32)
        self.pt._on_release(event)
        self.assertEqual(3, self.event_count, "Event must be triggered")
        self.assertEqual((31, 32), self.point_xy, "Event coordinate must be passed to callback")
        self.assertEqual(pt_point_bck, self.pt.point_xy, "point._on_release() must correctly set point_xy coordinates")

    def test_none_callback(self):
        self.pt = interactive_overlays.Point(canvas=self.canvas, point_xy=self.point_xy, label="ok", on_click=None, on_drag=None, on_release=None)

        binds = self.canvas.tag_bind(self.pt.circle_id)
        self.assertIn(EventName.CLICK, binds, f"{EventName.CLICK} tag not binded")
        self.assertIn(EventName.DRAG, binds, f"{EventName.DRAG} tag not binded")
        self.assertIn(EventName.RELEASE, binds, f"{EventName.RELEASE} tag not binded")
        self.assertIn(EventName.ENTER, binds, f"{EventName.ENTER} tag not binded")
        self.assertIn(EventName.LEAVE, binds, f"{EventName.LEAVE} tag not binded")

        self.assertEqual(0, self.event_count)
        self.assertEqual((0, 0), self.point_xy)

        event = Event(123, 456)
        self.pt._on_click(event)
        self.pt._on_drag(event)
        self.pt._on_release(event)

        self.assertEqual(0, self.event_count, "Event must not be triggered")
        self.assertEqual((0, 0), self.point_xy, "Event must not be triggered")
        self.assertEqual((123, 456), self.pt.point_xy, "point.point_xy must change even if no user callback is set")

    def test_event_sequence(self):
        self.pt = interactive_overlays.Point(canvas=self.canvas, point_xy=self.point_xy, label="ok", on_click=None, on_drag=None, on_release=None)
        event = Event(0, 0)

        self.assertEqual(self.pt.state, interactive_overlays.State.NORMAL)

        self.pt._on_enter(event)
        self.assertEqual(self.pt.state, interactive_overlays.State.HOVERED)

        self.pt._on_click(event)
        self.assertEqual(self.pt.state, interactive_overlays.State.DRAGGED)

        self.pt._on_drag(event)
        self.assertEqual(self.pt.state, interactive_overlays.State.DRAGGED)

        self.pt._on_release(event)
        self.assertEqual(self.pt.state, interactive_overlays.State.HOVERED)

        self.pt._on_leave(event)
        self.assertEqual(self.pt.state, interactive_overlays.State.NORMAL)

    def test_event_raise(self):
        self.pt = interactive_overlays.Point(
            canvas=self.canvas, point_xy=self.point_xy, label="ok", on_click=None, on_drag=self.on_event_error, on_release=None
        )
        event = Event(0, 0)

        with self.assertRaises(Exception):
            self.pt._on_drag(event)

    def test_visible(self):
        expected_state = {True: "normal", False: "hidden"}
        self.pt = interactive_overlays.Point(canvas=self.canvas, point_xy=self.point_xy, label="ok")

        self.pt.update()
        self.assertTrue(self.pt.visible, msg="Point must be visible by default")
        state = self.canvas.itemcget(self.pt.circle_id, "state")
        self.assertEqual(state, "normal", msg="Point must be visible by default")

        for val in [False, False, True, True]:
            self.pt.set_visible(val)
            self.pt.update()
            self.assertEqual(self.pt.visible, val, msg="set_visible function must work properly")
            state = self.canvas.itemcget(self.pt.circle_id, "state")
            self.assertEqual(state, expected_state[val], msg="set_visible function must work properly")

            self.assertEqual(self.pt.visible, val)




import numpy as np
import numpy.typing as npt
import cv2
import guibbon as gbn

Image_t = npt.NDArray[np.uint8]


class DemoMVCAppFail:
    """
    This demonstrate a BAD architecture. This is just an example for you to see the difference compared to a correct architecture
    The difference here is that the on_drag callback is calling update() and show() which is bad
    TODO: try it with a full cv2 widget (mouse event, or trackbar). And make sure the bug also appear. Otherwise a fix must be done
    """

    def __init__(self, filename):
        img = cv2.imread(filename)
        if img is None:
            raise FileNotFoundError(f"Image file not found: {filename}")
        self.img: Image_t = img.astype(np.uint8)
        self.winname = "demo app"
        gbn.namedWindow(self.winname)
        gbn.createInteractivePoint(self.winname, (100, 100), "point", on_drag=self.on_drag)

        self.x = 0
        self.y = 0

        self.res: Image_t
        self.update()

    def update(self):
        w, h = self.img.shape[:2]
        x = self.x / w
        y = self.y / h
        gain_b = max(0.5, min(2.0, 1.5 * x + 0.5))
        gain_r = max(0.5, min(2.0, 1.5 * y + 0.5))

        res = self.img.copy()
        res = res.astype(np.uint16)
        res[:, :, 0:1] = res[:, :, 0:1] * gain_b
        res[:, :, 2:3] = res[:, :, 2:3] * gain_r
        res[res > 255] = 255
        res = res.astype(np.uint8)
        self.res = res

    def show(self):
        gbn.imshow(self.winname, self.res)
        gbn.waitKeyEx(0)

    def on_drag(self, event):
        self.x = event.x
        self.y = event.y
        self.update()
        self.show()


if __name__ == "__main__":
    mc = DemoMVCAppFail("ressources/dog.jpg")
    mc.show()



import unittest
import guibbon as gbn
from guibbon.widgets.multi_slider_widget import MultiSliderWidget


class TestMultiSliderWidget(unittest.TestCase):
    def setUp(self):
        self.winname = "win0"
        gbn.namedWindow(self.winname)
        self.guibbon_instance = gbn.Guibbon.instances["win0"]
        self.callback_count = 0
        self.last_positions_values = None

    def tearDown(self):
        # Clean up
        if self.winname in gbn.Guibbon.instances:
            self.guibbon_instance.on_closing()

    def callback(self, positions_values):
        self.callback_count += 1
        self.last_positions_values = positions_values

    def test_create_multislider(self):
        """Test basic creation of multislider widget"""
        values = [0, 1, 2, 3, 4, 5]
        initial_positions = [1, 3]
        
        ms = gbn.create_multislider(
            winname=self.winname,
            multislider_name="Test MultiSlider",
            values=values,
            initial_indexes=initial_positions,
            on_drag=self.callback,
            on_release=self.callback
        )
        
        self.assertIsNotNone(ms)
        self.assertIsInstance(ms, MultiSliderWidget)
        
        # Check initial positions
        positions = ms.get_positions()
        self.assertEqual(positions, initial_positions)
        
        # Check values
        result_values = ms.get_values()
        self.assertEqual(result_values, values)

    def test_get_positions(self):
        """Test getting cursor positions"""
        values = list(range(10))
        initial_positions = [2, 5, 8]
        
        ms = gbn.create_multislider(
            winname=self.winname,
            multislider_name="Test",
            values=values,
            initial_indexes=initial_positions
        )
        
        positions = ms.get_positions()
        self.assertEqual(positions, initial_positions)
        self.assertEqual(len(positions), 3)

    def test_set_positions_without_callback(self):
        """Test setting positions without triggering callback"""
        values = list(range(10))
        initial_positions = [1, 5]
        
        ms = gbn.create_multislider(
            winname=self.winname,
            multislider_name="Test",
            values=values,
            initial_indexes=initial_positions,
            on_drag=self.callback
        )
        
        self.callback_count = 0
        new_positions = [3, 7]
        ms.set_positions(new_positions, trigger_callback=False)
        
        # Callback should not be triggered
        self.assertEqual(self.callback_count, 0)
        
        # Positions should be updated
        positions = ms.get_positions()
        self.assertEqual(positions, new_positions)

    def test_set_positions_with_callback(self):
        """Test setting positions with callback trigger"""
        values = list(range(10))
        initial_positions = [1, 5]
        
        ms = gbn.create_multislider(
            winname=self.winname,
            multislider_name="Test",
            values=values,
            initial_indexes=initial_positions,
            on_drag=self.callback,
            on_release=self.callback
        )
        
        self.callback_count = 0
        new_positions = [2, 6]
        ms.set_positions(new_positions, trigger_callback=True)
        
        # Both on_drag and on_release callbacks should be triggered
        self.assertEqual(self.callback_count, 2)
        
        # Check the callback received correct data
        self.assertIsNotNone(self.last_positions_values)
        expected = [(2, values[2]), (6, values[6])]
        self.assertEqual(self.last_positions_values, expected)

    def test_get_values(self):
        """Test getting values list"""
        values = ['a', 'b', 'c', 'd', 'e']
        initial_positions = [0, 2]
        
        ms = gbn.create_multislider(
            winname=self.winname,
            multislider_name="Test",
            values=values,
            initial_indexes=initial_positions
        )
        
        result_values = ms.get_values()
        self.assertEqual(result_values, values)
        # Verify it's a copy, not the same list
        self.assertIsNot(result_values, values)

    def test_set_values_without_callback(self):
        """Test setting new values without triggering callback"""
        initial_values = list(range(5))
        initial_positions = [1, 3]
        
        ms = gbn.create_multislider(
            winname=self.winname,
            multislider_name="Test",
            values=initial_values,
            initial_indexes=initial_positions,
            on_drag=self.callback
        )
        
        self.callback_count = 0
        new_values = list(range(10, 20))
        ms.set_values(new_values, trigger_callback=False)
        
        # Callback should not be triggered
        self.assertEqual(self.callback_count, 0)
        
        # Values should be updated
        result_values = ms.get_values()
        self.assertEqual(result_values, new_values)

    def test_set_values_with_callback(self):
        """Test setting new values with callback trigger"""
        initial_values = list(range(5))
        initial_positions = [1, 3]
        
        ms = gbn.create_multislider(
            winname=self.winname,
            multislider_name="Test",
            values=initial_values,
            initial_indexes=initial_positions,
            on_drag=self.callback,
            on_release=self.callback
        )
        
        self.callback_count = 0
        new_values = list(range(10, 20))
        ms.set_values(new_values, trigger_callback=True)
        
        # Both callbacks should be triggered
        self.assertEqual(self.callback_count, 2)

    def test_add_cursor(self):
        """Test adding a cursor dynamically"""
        values = list(range(10))
        initial_positions = [2]
        
        ms = gbn.create_multislider(
            winname=self.winname,
            multislider_name="Test",
            values=values,
            initial_indexes=initial_positions
        )
        
        # Initially should have 1 cursor
        self.assertEqual(len(ms.get_positions()), 1)
        
        # Add a cursor at position 5
        ms.add_cursor(position=5)
        
        # Should now have 2 cursors
        positions = ms.get_positions()
        self.assertEqual(len(positions), 2)
        self.assertIn(5, positions)

    def test_remove_cursor(self):
        """Test removing a cursor"""
        values = list(range(10))
        initial_positions = [2, 5, 8]
        
        ms = gbn.create_multislider(
            winname=self.winname,
            multislider_name="Test",
            values=values,
            initial_indexes=initial_positions
        )
        
        # Initially should have 3 cursors
        self.assertEqual(len(ms.get_positions()), 3)
        
        # Remove a cursor
        ms.remove_cursor()
        
        # Should now have 2 cursors
        positions = ms.get_positions()
        self.assertEqual(len(positions), 2)

    def test_remove_cursor_minimum(self):
        """Test that at least one cursor remains"""
        values = list(range(10))
        initial_positions = [5]
        
        ms = gbn.create_multislider(
            winname=self.winname,
            multislider_name="Test",
            values=values,
            initial_indexes=initial_positions
        )
        
        # Try to remove the only cursor
        ms.remove_cursor()
        
        # Should still have 1 cursor (minimum)
        positions = ms.get_positions()
        self.assertEqual(len(positions), 1)

    def test_positions_adjust_when_values_shrink(self):
        """Test that positions are adjusted when value list shrinks"""
        initial_values = list(range(10))
        initial_positions = [3, 7, 9]
        
        ms = gbn.create_multislider(
            winname=self.winname,
            multislider_name="Test",
            values=initial_values,
            initial_indexes=initial_positions
        )
        
        # Set shorter value list
        new_values = list(range(5))
        ms.set_values(new_values, trigger_callback=False)
        
        # Positions beyond the new max should be clamped
        positions = ms.get_positions()
        for pos in positions:
            self.assertLessEqual(pos, len(new_values) - 1)

    def test_set_positions_adds_cursors(self):
        """Test that set_positions can add cursors if needed"""
        values = list(range(10))
        initial_positions = [2]
        
        ms = gbn.create_multislider(
            winname=self.winname,
            multislider_name="Test",
            values=values,
            initial_indexes=initial_positions
        )
        
        # Set more positions than current cursors
        new_positions = [1, 3, 5, 7]
        ms.set_positions(new_positions, trigger_callback=False)
        
        # Should now have 4 cursors
        positions = ms.get_positions()
        self.assertEqual(len(positions), 4)
        self.assertEqual(positions, new_positions)

    def test_set_positions_removes_cursors(self):
        """Test that set_positions can remove cursors if needed"""
        values = list(range(10))
        initial_positions = [1, 3, 5, 7]
        
        ms = gbn.create_multislider(
            winname=self.winname,
            multislider_name="Test",
            values=values,
            initial_indexes=initial_positions
        )
        
        # Set fewer positions than current cursors
        new_positions = [2, 6]
        ms.set_positions(new_positions, trigger_callback=False)
        
        # Should now have 2 cursors
        positions = ms.get_positions()
        self.assertEqual(len(positions), 2)
        self.assertEqual(positions, new_positions)

    def test_callback_parameters(self):
        """Test that callbacks receive correct parameters"""
        values = ['apple', 'banana', 'cherry', 'date']
        initial_positions = [0, 2]
        
        ms = gbn.create_multislider(
            winname=self.winname,
            multislider_name="Test",
            values=values,
            initial_indexes=initial_positions,
            on_drag=self.callback
        )
        
        # Trigger callback by setting positions
        new_positions = [1, 3]
        ms.set_positions(new_positions, trigger_callback=True)
        
        # Check callback received correct format: list of (position, value) tuples
        expected = [(1, 'banana'), (3, 'date')]
        self.assertEqual(self.last_positions_values, expected)

    def test_none_callbacks(self):
        """Test that widget works with None callbacks"""
        values = list(range(5))
        initial_positions = [1, 3]
        
        # Should not raise exception
        ms = gbn.create_multislider(
            winname=self.winname,
            multislider_name="Test",
            values=values,
            initial_indexes=initial_positions,
            on_drag=None,
            on_release=None
        )
        
        # Operations should work without callbacks
        ms.set_positions([0, 4], trigger_callback=True)
        ms.add_cursor(2)
        ms.remove_cursor()
        
        # Should complete without errors
        self.assertEqual(len(ms.get_positions()), 2)


if __name__ == "__main__":
    unittest.main()



import tkinter as tk
from typing import Callable, Any, Sequence, Optional

from guibbon.typedef import Point2Di
from .base import Point

MultiSliderState = list[tuple[int, Any]]
CallbackMultiSlider = Callable[[MultiSliderState], None]


class MultiSliderOverlay:
    colors = {
        "grey": "#%02x%02x%02x" % (191, 191, 191),
    }

    def __init__(
        self,
        canvas: tk.Canvas,
        values: Sequence[Any],
        initial_positions: Sequence[int],
        on_drag: Optional[CallbackMultiSlider] = None,
        on_release: Optional[CallbackMultiSlider] = None,
    ):
        self.canvas = canvas
        self.values: list[Any] = [v for v in values]

        self.on_drag = on_drag
        self.on_release = on_release

        self.line_ids = []
        self.line_ids.append(self.canvas.create_line(-1, -1, -1, -1, fill=MultiSliderOverlay.colors["grey"], width=3))
        for i in range(len(values)):
            self.line_ids.append(self.canvas.create_line(-1, -1, -1, -1, fill=MultiSliderOverlay.colors["grey"], width=3))

        # self.positions_values: MultiSliderState = []
        self.cursors: list[Point] = []

        self.cursor_positions: list[int] = []
        for k in range(len(initial_positions)):
            self.add_cursor(position=initial_positions[k])

        self.update_canvas()

    def add_cursor(self, position: int = 0):
        new_cursor_id = len(self.cursors)

        def on_drag_wrap(event: tk.Event) -> None:
            self.on_drag_callback(new_cursor_id, event)
            return None

        cursor = Point(
            canvas=self.canvas,
            point_xy=(-1, -1),
            on_drag=on_drag_wrap,
            on_release=None if self.on_release is None else self.on_release_callback,
        )
        self.cursors.append(cursor)
        self.cursor_positions.append(position)
        self.update_canvas()

    def remove_cursor(self):
        if len(self.cursors) <= 1:
            return
        cursor = self.cursors.pop()
        cursor.set_visible(False)
        cursor.delete()
        self.cursor_positions.pop()
        self.update_canvas()

    def update_canvas(self):
        self.canvas.update_idletasks()  # forces to compute rendered positions_values and size of the canvas

        N = len(self.values)

        # draw ticks
        for i, line_id in enumerate(self.line_ids[:-1]):
            line_x, line_y = self.slider2canvas((i, 0))
            self.canvas.coords(line_id, line_x, line_y - 2, line_x, line_y + 3)
            self.canvas.itemconfig(line_id, state="normal")
            self.canvas.tag_raise(line_id)

        # draw horizontal line
        line_x0, line_y0 = self.slider2canvas((0, 0))
        line_x1, line_y1 = self.slider2canvas((N - 1, 0))
        self.canvas.coords(self.line_ids[-1], line_x0, line_y0, line_x1, line_y1)
        self.canvas.itemconfig(self.line_ids[-1], state="normal")
        self.canvas.tag_raise(self.line_ids[-1])

        # draw cursors
        for cursor_pos, cursor in zip(self.cursor_positions, self.cursors):
            x_slider = cursor_pos
            point_xy_can = self.slider2canvas((x_slider, 0))
            cursor.set_can_point_xy(point_xy_can)
            cursor.update()

    def slider2canvas(self, point_xy: Point2Di) -> Point2Di:
        w_can = self.canvas.winfo_width()
        h_can = self.canvas.winfo_height()
        margin = h_can // 2

        w_slider = max(1, len(self.values) - 1)
        w_slider_can = max(1, w_can - 2 * margin)
        scale = w_slider / w_slider_can

        x_slider, y_slider = point_xy

        x_can = int(round(x_slider / scale + margin))
        y_can = margin + 1
        return x_can, y_can

    def canvas2slider(self, point_xy: Point2Di) -> Point2Di:
        w_can = self.canvas.winfo_width()
        h_can = self.canvas.winfo_height()
        margin = h_can // 2

        w_slider = max(1, len(self.values) - 1)
        w_slider_can = max(1, w_can - 2 * margin)
        scale = w_slider / w_slider_can

        x_can, y_can = point_xy
        x_can_min = margin
        x_slider = (x_can - x_can_min) * scale
        x_slider = min(max(0.0, x_slider), w_slider)
        x_slider = int(round(x_slider))

        y_slider = 0
        return x_slider, y_slider

    def on_drag_callback(self, cursor_id, event: tk.Event):
        x_slider, y_slider = self.canvas2slider((event.x, event.y))
        x_can, y_can = self.slider2canvas((x_slider, y_slider))
        self.cursors[cursor_id].set_can_point_xy((x_can, y_can))
        self.cursors[cursor_id].update()

        old_ind = self.cursor_positions[cursor_id]
        new_ind = x_slider
        if old_ind == new_ind:
            return

        self.cursor_positions[cursor_id] = x_slider

        if self.on_drag is not None:
            positions_values = [(pos, self.values[pos]) for pos in self.cursor_positions]
            self.on_drag(positions_values)

    def on_release_callback(self, event: tk.Event):
        if self.on_release is not None:
            positions_values = [(pos, self.values[pos]) for pos in self.cursor_positions]
            self.on_release(positions_values)

    def get_positions(self) -> list[int]:
        return self.cursor_positions + []

    def set_positions(self, positions: Sequence[int], trigger_callback=True):
        delta = len(positions) - len(self.cursor_positions)
        if delta > 0:
            for _ in range(delta):
                self.add_cursor(position=0)
        elif delta < 0:
            for _ in range(-delta):
                self.remove_cursor()

        self.cursor_positions = [pos for pos in positions]
        self.update_canvas()
        if trigger_callback:
            positions_values = [(pos, self.values[pos]) for pos in self.cursor_positions]
            if self.on_drag is not None:
                self.on_drag(positions_values)
            if self.on_release is not None:
                self.on_release(positions_values)

    def get_values(self) -> list[Any]:
        return self.values + []

    def set_values(self, values: Sequence[Any], trigger_callback=True):
        self.values = [v for v in values]
        self.cursor_positions = [min(pos, len(self.values) - 1) for pos in self.cursor_positions]
        self.update_canvas()
        if trigger_callback:
            positions_values = [(pos, self.values[pos]) for pos in self.cursor_positions]
            if self.on_drag is not None:
                self.on_drag(positions_values)
            if self.on_release is not None:
                self.on_release(positions_values)



import tkinter as tk
from tkinter import ttk
from typing import Callable, Any

CallbackTreeview = Callable[[list[str], Any], None]


class TreeNode:
    def __init__(self, name="", value=None):
        self.name = name
        self.value = value
        self.children = {}

    def set(self, path, value=None):
        if len(path) == 0:
            self.value = value
            return
        child_name = path[0]
        if child_name not in self.children.keys():
            self.children[child_name] = TreeNode(child_name)
        self.children[child_name].set(path[1:], value)

    def get(self, path):
        if len(path) == 0:
            return self
        child_name = path[0]
        return self.children[child_name].get(path[1:])

    def str(self, indent=0):
        lines = []
        data_str = "" if self.value is None else str(self.value)
        lines.append(f"{self.name}: {data_str}")
        for child in self.children.values():
            indent_str = "  : " * indent + "  + "
            lines.append(indent_str + child.str(indent + 1))
        return "\n".join(lines)

    def __str__(self):
        return self.str()

    def depth_first_walk(self):
        yield [self.name], self.value, self
        for child in self.children.values():
            for subpath, subname, subchild in child.depth_first_walk():
                yield [self.name] + subpath, subname, subchild


class TreeviewWidget:
    def __init__(self, tk_frame, name: str, tree: TreeNode, on_click: CallbackTreeview):
        self.tk_frame = tk_frame
        self.name = name
        self.tree: TreeNode = tree
        self.on_click = on_click
        self.frame = tk.Frame(self.tk_frame)
        tk.Label(self.frame, text=self.name).pack(padx=2, side=tk.TOP, anchor=tk.W)
        self.tk_treeview = ttk.Treeview(self.frame)

        self.build_tree("", self.tree)

        self.tk_treeview.bind("<Button-1>", self.callback)

        self.tk_treeview.pack(padx=2, side=tk.TOP, anchor=tk.W)
        self.frame.pack(padx=2, side=tk.TOP, anchor=tk.W)

    def build_tree(self, parent, tree_node):
        assert isinstance(tree_node, TreeNode)
        for subtree_node in tree_node.children.values():
            item = self.tk_treeview.insert(parent, tk.END, text=subtree_node.name)
            # self.tk_treeview.tag_bind(tk_node, "<Button-1>", self.callback)
            self.build_tree(item, subtree_node)

    def callback(self, event):
        try:
            item = self.tk_treeview.identify('item', event.x, event.y)
            names: list[str] = []
            while item != "":
                names = [self.tk_treeview.item(item, "text")] + names
                item = self.tk_treeview.parent(item)

            value = self.tree.get(names)
            self.on_click(names, value)
            # item = self.tk_treeview.selection()[0]
            # print("you clicked on", self.tk_treeview.item(item, "text"))
        except Exception:
            print("Fail")


def main():
    tree = TreeNode("master")
    tree.set(["mon", "petit", "node"], 12)
    tree.set(["mon", "petit", "node", "gentil"], 56)
    tree.set(["mon", "petit", "arbre"], [34])
    petit = tree.get(["mon", "petit", "arbre"])
    petit.value.append(3344)
    print(tree)

    for node in tree.depth_first_walk():
        print(node)


if __name__ == '__main__':
    main()



import tkinter as tk
from typing import Callable

CallbackButton = Callable[[], None]


class ButtonWidget:
    def __init__(self, tk_frame, text, on_click: CallbackButton):
        self.tk_frame = tk_frame
        self.button = tk.Button(tk_frame, text=text, command=on_click)
        self._is_visible = False
        self.set_visible(True)

    def set_visible(self, val):
        if val == self._is_visible:
            return
        self._is_visible = val

        if val:
            self.button.pack(side=tk.LEFT)
        else:
            self.button.pack_forget()




from .base import Point as Point, Magnets as Magnets, State as State
from .polygons import Polygon as Polygon, Rectangle as Rectangle
from .multi_slider import MultiSliderOverlay as MultiSliderOverlay, MultiSliderState as MultiSliderState, CallbackMultiSlider as CallbackMultiSlider